import copy
import logging
import math
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
import scipy.stats
import statsmodels.api as sm
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from tqdm.auto import tqdm

from dowhy.causal_estimator import CausalEstimate, CausalEstimator
from dowhy.causal_estimators.econml import Econml
from dowhy.causal_estimators.linear_regression_estimator import LinearRegressionEstimator
from dowhy.causal_estimators.regression_estimator import RegressionEstimator
from dowhy.causal_identifier.identified_estimand import IdentifiedEstimand
from dowhy.causal_refuter import CausalRefutation, CausalRefuter, choose_variables
from dowhy.causal_refuters.evalue_sensitivity_analyzer import EValueSensitivityAnalyzer
from dowhy.causal_refuters.linear_sensitivity_analyzer import LinearSensitivityAnalyzer
from dowhy.causal_refuters.non_parametric_sensitivity_analyzer import NonParametricSensitivityAnalyzer
from dowhy.causal_refuters.partial_linear_sensitivity_analyzer import PartialLinearSensitivityAnalyzer

logger = logging.getLogger(__name__)


DEFAULT_CONVERGENCE_THRESHOLD = 0.1
DEFAULT_C_STAR_MAX = 1000


class AddUnobservedCommonCause(CausalRefuter):

    """Add an unobserved confounder for refutation.

    AddUnobservedCommonCause class supports three methods:
        1) Simulation of an unobserved confounder
        2) Linear partial R2 : Sensitivity Analysis for linear models.
        3) Non-Parametric partial R2 based : Sensitivity Analyis for non-parametric models.

    Supports additional parameters that can be specified in the refute_estimate() method.
    """

    def __init__(self, *args, **kwargs):
        """
        Initialize the parameters required for the refuter.

        For direct_simulation, if effect_strength_on_treatment or effect_strength_on_outcome is not
        given, it is calculated automatically as a range between the
        minimum and maximum effect strength of observed confounders on treatment
        and outcome respectively.

        :param simulation_method: The method to use for simulating effect of unobserved confounder. Possible values are ["direct-simulation", "linear-partial-R2", "non-parametric-partial-R2", "e-value"].
        :param confounders_effect_on_treatment: str : The type of effect on the treatment due to the unobserved confounder. Possible values are ['binary_flip', 'linear']
        :param confounders_effect_on_outcome: str : The type of effect on the outcome due to the unobserved confounder. Possible values are ['binary_flip', 'linear']
        :param effect_strength_on_treatment: float, numpy.ndarray: [Used when simulation_method="direct-simulation"] Strength of the confounder's effect on treatment. When confounders_effect_on_treatment is linear,  it is the regression coefficient. When the confounders_effect_on_treatment is binary flip, it is the probability with which effect of unobserved confounder can invert the value of the treatment.
        :param effect_strength_on_outcome: float, numpy.ndarray: Strength of the confounder's effect on outcome. Its interpretation depends on confounders_effect_on_outcome and the simulation_method. When simulation_method is direct-simulation, for a linear effect it behaves like the regression coefficient and for a binary flip, it is the probability with which it can invert the value of the outcome.
        :param partial_r2_confounder_treatment: float, numpy.ndarray: [Used when simulation_method is linear-partial-R2 or non-parametric-partial-R2] Partial R2 of the unobserved confounder wrt the treatment conditioned on the observed confounders. Only in the case of general non-parametric-partial-R2, it is the fraction of variance in the reisz representer that is explained by the unobserved confounder; specifically  (1-r), where r is the ratio of variance of reisz representer, alpha^2, based on observed confounders and that based on all confounders.
        :param partial_r2_confounder_outcome: float, numpy.ndarray: [Used when simulation_method is linear-partial-R2 or non-parametric-partial-R2] Partial R2 of the unobserved confounder wrt the outcome conditioned on the treatment and observed confounders.
        :param frac_strength_treatment: float: This parameter decides the effect strength of the simulated confounder as a fraction of the effect strength of observed confounders on treatment. Defaults to 1.
        :param frac_strength_outcome: float: This parameter decides the effect strength of the simulated confounder as a fraction of the effect strength of observed confounders on outcome. Defaults to 1.
        :param plotmethod: string: Type of plot to be shown. If None, no plot is generated. This parameter is used only only when more than one treatment confounder effect values or outcome confounder effect values are provided. Default is "colormesh". Supported values are "contour", "colormesh" when more than one value is provided for both confounder effect value parameters; "line" when provided for only one of them.
        :param percent_change_estimate: It is the percentage of reduction of treatment estimate that could alter the results (default = 1).
                                        if percent_change_estimate = 1, the robustness value describes the strength of association of confounders with treatment and outcome in order to reduce the estimate by 100% i.e bring it down to 0. (relevant only for Linear Sensitivity Analysis, ignore for rest)
        :param confounder_increases_estimate: True implies that confounder increases the absolute value of estimate and vice versa. (Default = False). (relevant only for Linear Sensitivity Analysis, ignore for rest)
        :param benchmark_common_causes: names of variables for bounding strength of confounders. (relevant only for partial-r2 based simulation methods)
        :param significance_level: confidence interval for statistical inference(default = 0.05). (relevant only for partial-r2 based simulation methods)
        :param null_hypothesis_effect: assumed effect under the null hypothesis. (relevant only for linear-partial-R2, ignore for rest)
        :param plot_estimate: Generate contour plot for estimate while performing sensitivity analysis. (default = True).
                              (relevant only for partial-r2 based simulation methods)
        :param num_splits: number of splits for cross validation. (default = 5). (relevant only for non-parametric-partial-R2 simulation method)
        :param shuffle_data : shuffle data or not before splitting into folds (default = False). (relevant only for non-parametric-partial-R2 simulation method)
        :param shuffle_random_seed: seed for randomly shuffling data. (relevant only for non-parametric-partial-R2 simulation method)
        :param alpha_s_estimator_param_list: list of dictionaries with parameters for finding alpha_s. (relevant only for non-parametric-partial-R2 simulation method)
        :param g_s_estimator_list: list of estimator objects for finding g_s. These objects should have fit() and predict() functions implemented. (relevant only for non-parametric-partial-R2 simulation method)
        :param g_s_estimator_param_list: list of dictionaries with parameters for tuning respective estimators in "g_s_estimator_list". The order of the dictionaries in the list should be consistent with the estimator objects order in "g_s_estimator_list". (relevant only for non-parametric-partial-R2 simulation method)
        """
        super().__init__(*args, **kwargs)
        self.simulation_method = kwargs["simulation_method"] if "simulation_method" in kwargs else "direct-simulation"
        self.effect_on_t = (
            kwargs["confounders_effect_on_treatment"] if "confounders_effect_on_treatment" in kwargs else "binary_flip"
        )
        self.effect_on_y = (
            kwargs["confounders_effect_on_outcome"] if "confounders_effect_on_outcome" in kwargs else "linear"
        )
        if self.simulation_method == "direct-simulation":
            self.kappa_t = kwargs["effect_strength_on_treatment"] if "effect_strength_on_treatment" in kwargs else None
            self.kappa_y = kwargs["effect_strength_on_outcome"] if "effect_strength_on_outcome" in kwargs else None
        elif self.simulation_method in ["linear-partial-R2", "non-parametric-partial-R2"]:
            self.kappa_t = (
                kwargs["partial_r2_confounder_treatment"] if "partial_r2_confounder_treatment" in kwargs else None
            )
            self.kappa_y = (
                kwargs["partial_r2_confounder_outcome"] if "partial_r2_confounder_outcome" in kwargs else None
            )
        elif self.simulation_method == "e-value":
            pass
        else:
            raise ValueError(
                "simulation method is not supported. Try direct-simulation, linear-partial-R2, non-parametric-partial-R2, or e-value"
            )
        self.frac_strength_treatment = (
            kwargs["effect_fraction_on_treatment"] if "effect_fraction_on_treatment" in kwargs else 1
        )
        self.frac_strength_outcome = (
            kwargs["effect_fraction_on_outcome"] if "effect_fraction_on_outcome" in kwargs else 1
        )

        self.plotmethod = kwargs["plotmethod"] if "plotmethod" in kwargs else "colormesh"
        self.percent_change_estimate = kwargs["percent_change_estimate"] if "percent_change_estimate" in kwargs else 1.0
        self.significance_level = kwargs["significance_level"] if "significance_level" in kwargs else 0.05
        self.confounder_increases_estimate = (
            kwargs["confounder_increases_estimate"] if "confounder_increases_estimate" in kwargs else False
        )
        self.benchmark_common_causes = (
            kwargs["benchmark_common_causes"] if "benchmark_common_causes" in kwargs else None
        )
        self.null_hypothesis_effect = kwargs["null_hypothesis_effect"] if "null_hypothesis_effect" in kwargs else 0
        self.plot_estimate = kwargs["plot_estimate"] if "plot_estimate" in kwargs else True
        self.num_splits = kwargs["num_splits"] if "num_splits" in kwargs else 5
        self.shuffle_data = kwargs["shuffle_data"] if "shuffle_data" in kwargs else False
        self.shuffle_random_seed = kwargs["shuffle_random_seed"] if "shuffle_random_seed" in kwargs else None
        self.alpha_s_estimator_param_list = (
            kwargs["alpha_s_estimator_param_list"] if "alpha_s_estimator_param_list" in kwargs else None
        )
        self.alpha_s_estimator_list = kwargs["alpha_s_estimator_list"] if "alpha_s_estimator_list" in kwargs else None
        self.g_s_estimator_list = kwargs["g_s_estimator_list"] if "g_s_estimator_list" in kwargs else None
        self.g_s_estimator_param_list = (
            kwargs["g_s_estimator_param_list"] if "g_s_estimator_param_list" in kwargs else None
        )
        self.plugin_reisz = kwargs["plugin_reisz"] if "plugin_reisz" in kwargs else False
        self.logger = logging.getLogger(__name__)

    def refute_estimate(self, show_progress_bar=False):
        if self.simulation_method == "linear-partial-R2":
            return sensitivity_linear_partial_r2(
                self._data,
                self._estimate,
                self._treatment_name,
                self.frac_strength_treatment,
                self.frac_strength_outcome,
                self.percent_change_estimate,
                self.benchmark_common_causes,
                self.significance_level,
                self.null_hypothesis_effect,
                self.plot_estimate,
            )
        elif self.simulation_method == "non-parametric-partial-R2":
            return sensitivity_non_parametric_partial_r2(
                self._estimate,
                self.kappa_t,
                self.kappa_y,
                self.frac_strength_treatment,
                self.frac_strength_outcome,
                self.benchmark_common_causes,
                self.plot_estimate,
                self.alpha_s_estimator_list,
                self.alpha_s_estimator_param_list,
                self.g_s_estimator_list,
                self.g_s_estimator_param_list,
                self.plugin_reisz,
            )
        elif self.simulation_method == "e-value":
            return sensitivity_e_value(
                self._data,
                self._target_estimand,
                self._estimate,
                self._treatment_name,
                self._outcome_name,
                self.plot_estimate,
            )
        elif self.simulation_method == "direct-simulation":
            refute = sensitivity_simulation(
                self._data,
                self._target_estimand,
                self._estimate,
                self._treatment_name,
                self._outcome_name,
                self.kappa_t,
                self.kappa_y,
                self.effect_on_t,
                self.effect_on_y,
                self.frac_strength_treatment,
                self.frac_strength_outcome,
                self.plotmethod,
                show_progress_bar,
            )
            refute.add_refuter(self)
            return refute

    def include_simulated_confounder(
        self, convergence_threshold=DEFAULT_CONVERGENCE_THRESHOLD, c_star_max=DEFAULT_C_STAR_MAX
    ):
        return include_simulated_confounder(
            self._data,
            self._treatment_name,
            self._outcome_name,
            self.kappa_t,
            self.kappa_y,
            self._variables_of_interest,
            convergence_threshold,
            c_star_max,
        )


def _infer_default_kappa_t(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    treatment_name: List[str],
    effect_on_t: str,
    frac_strength_treatment: float,
    len_kappa_t: int = 10,
):
    """Infer default effect strength of simulated confounder on treatment."""
    observed_common_causes_names = target_estimand.get_backdoor_variables()
    if len(observed_common_causes_names) > 0:
        observed_common_causes = data[observed_common_causes_names]
        observed_common_causes = pd.get_dummies(observed_common_causes, drop_first=True)
    else:
        raise ValueError(
            "There needs to be at least one common cause to"
            + "automatically compute the default value of kappa_t."
            + " Provide a value for kappa_t"
        )
    t = data[treatment_name]
    # Standardizing the data
    observed_common_causes = StandardScaler().fit_transform(observed_common_causes)
    if effect_on_t == "binary_flip":
        # Fit a model containing all confounders and compare predictions
        # using all features compared to all features except a given
        # confounder.
        tmodel = LogisticRegression().fit(observed_common_causes, t)
        tpred = tmodel.predict(observed_common_causes).astype(int)
        flips = []
        for i in range(observed_common_causes.shape[1]):
            oldval = np.copy(observed_common_causes[:, i])
            observed_common_causes[:, i] = 0
            tcap = tmodel.predict(observed_common_causes).astype(int)
            observed_common_causes[:, i] = oldval
            flips.append(np.sum(abs(tcap - tpred)) / tpred.shape[0])
        min_coeff, max_coeff = min(flips), max(flips)
    elif effect_on_t == "linear":
        # Estimating the regression coefficient from standardized features to t
        corrcoef_var_t = np.corrcoef(observed_common_causes, t, rowvar=False)[-1, :-1]
        std_dev_t = np.std(t)[0]
        max_coeff = max(corrcoef_var_t) * std_dev_t
        min_coeff = min(corrcoef_var_t) * std_dev_t
    else:
        raise NotImplementedError("'" + effect_on_t + "' method not supported for confounders' effect on treatment")

    min_coeff, max_coeff = _compute_min_max_coeff(min_coeff, max_coeff, frac_strength_treatment)
    # By default, return a plot with 10 points
    # consider 10 values of the effect of the unobserved confounder
    step = (max_coeff - min_coeff) / len_kappa_t
    logger.info("(Min, Max) kappa_t for observed common causes, ({0}, {1})".format(min_coeff, max_coeff))
    if np.equal(max_coeff, min_coeff):
        return max_coeff
    else:
        return np.arange(min_coeff, max_coeff, step)


def _compute_min_max_coeff(min_coeff: float, max_coeff: float, effect_strength_fraction: np.ndarray):
    max_coeff = effect_strength_fraction * max_coeff
    min_coeff = effect_strength_fraction * min_coeff
    return min_coeff, max_coeff


def _infer_default_kappa_y(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    outcome_name: List[str],
    effect_on_y: str,
    frac_strength_outcome: float,
    len_kappa_y: int = 10,
):
    """Infer default effect strength of simulated confounder on treatment."""
    observed_common_causes_names = target_estimand.get_backdoor_variables()
    if len(observed_common_causes_names) > 0:
        observed_common_causes = data[observed_common_causes_names]
        observed_common_causes = pd.get_dummies(observed_common_causes, drop_first=True)
    else:
        raise ValueError(
            "There needs to be at least one common cause to"
            + "automatically compute the default value of kappa_y."
            + " Provide a value for kappa_y"
        )
    y = data[outcome_name]
    # Standardizing the data
    observed_common_causes = StandardScaler().fit_transform(observed_common_causes)
    if effect_on_y == "binary_flip":
        # Fit a model containing all confounders and compare predictions
        # using all features compared to all features except a given
        # confounder.
        ymodel = LogisticRegression().fit(observed_common_causes, y)
        ypred = ymodel.predict(observed_common_causes).astype(int)
        flips = []
        for i in range(observed_common_causes.shape[1]):
            oldval = np.copy(observed_common_causes[:, i])
            observed_common_causes[:, i] = 0
            ycap = ymodel.predict(observed_common_causes).astype(int)
            observed_common_causes[:, i] = oldval
            flips.append(np.sum(abs(ycap - ypred)) / ypred.shape[0])
        min_coeff, max_coeff = min(flips), max(flips)
    elif effect_on_y == "linear":
        corrcoef_var_y = np.corrcoef(observed_common_causes, y, rowvar=False)[-1, :-1]
        std_dev_y = np.std(y)[0]
        max_coeff = max(corrcoef_var_y) * std_dev_y
        min_coeff = min(corrcoef_var_y) * std_dev_y
    else:
        raise NotImplementedError("'" + effect_on_y + "' method not supported for confounders' effect on outcome")
    min_coeff, max_coeff = _compute_min_max_coeff(min_coeff, max_coeff, frac_strength_outcome)
    # By default, return a plot with 10 points
    # consider 10 values of the effect of the unobserved confounder
    step = (max_coeff - min_coeff) / len_kappa_y
    logger.info("(Min, Max) kappa_y for observed common causes, ({0}, {1})".format(min_coeff, max_coeff))
    if np.equal(max_coeff, min_coeff):
        return max_coeff
    else:
        return np.arange(min_coeff, max_coeff, step)


def _include_confounders_effect(
    data: pd.DataFrame,
    new_data: pd.DataFrame,
    effect_on_t: str,
    treatment_name: str,
    kappa_t: float,
    effect_on_y: str,
    outcome_name: str,
    kappa_y: float,
):
    """
    This function deals with the change in the value of the data due to the effect of the unobserved confounder.
    In the case of a binary flip, we flip only if the random number is greater than the threshold set.
    In the case of a linear effect, we use the variable as the linear regression constant.

    :param new_data: pandas.DataFrame: The data to be changed due to the effects of the unobserved confounder.
    :param kappa_t: numpy.float64: The value of the threshold for binary_flip or the value of the regression coefficient for linear effect.
    :param kappa_y: numpy.float64: The value of the threshold for binary_flip or the value of the regression coefficient for linear effect.

    :return: pandas.DataFrame: The DataFrame that includes the effects of the unobserved confounder.
    """
    num_rows = data.shape[0]
    stdnorm = scipy.stats.norm()
    w_random = stdnorm.rvs(num_rows)

    if effect_on_t == "binary_flip":
        alpha = 2 * kappa_t - 1 if kappa_t >= 0.5 else 1 - 2 * kappa_t
        interval = stdnorm.interval(alpha)
        rel_interval = interval[0] if kappa_t >= 0.5 else interval[1]
        new_data.loc[rel_interval <= w_random, treatment_name] = (
            1 - new_data.loc[rel_interval <= w_random, treatment_name]
        )
        for tname in treatment_name:
            if pd.api.types.is_bool_dtype(data[tname]):
                new_data = new_data.astype({tname: "bool"}, copy=False)
    elif effect_on_t == "linear":
        confounder_t_effect = kappa_t * w_random
        # By default, we add the effect of simulated confounder for treatment.
        # But subtract it from outcome to create a negative correlation
        # assuming that the original confounder's effect was positive on both.
        # This is to remove the effect of the original confounder.
        new_data[treatment_name] = new_data[treatment_name].values + np.ndarray(
            shape=(num_rows, 1), buffer=confounder_t_effect
        )
    else:
        raise NotImplementedError("'" + effect_on_t + "' method not supported for confounders' effect on treatment")

    if effect_on_y == "binary_flip":
        alpha = 2 * kappa_y - 1 if kappa_y >= 0.5 else 1 - 2 * kappa_y
        interval = stdnorm.interval(alpha)
        rel_interval = interval[0] if kappa_y >= 0.5 else interval[1]
        new_data.loc[rel_interval <= w_random, outcome_name] = 1 - new_data.loc[rel_interval <= w_random, outcome_name]
        for yname in outcome_name:
            if pd.api.types.is_bool_dtype(data[yname]):
                new_data = new_data.astype({yname: "bool"}, copy=False)
    elif effect_on_y == "linear":
        confounder_y_effect = (-1) * kappa_y * w_random
        # By default, we add the effect of simulated confounder for treatment.
        # But subtract it from outcome to create a negative correlation
        # assuming that the original confounder's effect was positive on both.
        # This is to remove the effect of the original confounder.
        new_data[outcome_name] = new_data[outcome_name].values + np.ndarray(
            shape=(num_rows, 1), buffer=confounder_y_effect
        )
    else:
        raise NotImplementedError("'" + effect_on_y + "' method not supported for confounders' effect on outcome")
    return new_data


def include_simulated_confounder(
    data: pd.DataFrame,
    treatment_name: str,
    outcome_name: str,
    kappa_t: float,
    kappa_y: float,
    variables_of_interest: List,
    convergence_threshold: float = DEFAULT_CONVERGENCE_THRESHOLD,
    c_star_max: int = DEFAULT_C_STAR_MAX,
):
    """
    This function simulates an unobserved confounder based on the data using the following steps:
    1. It calculates the "residuals"  from the treatment and outcome model
    i.) The outcome model has outcome as the dependent variable and all the observed variables including treatment as independent variables
    ii.) The treatment model has treatment as the dependent variable and all the observed variables as independent variables.

    2. U is an intermediate random variable drawn from the normal distribution with the weighted average of residuals as mean and a unit variance
    U ~ N(c1*d_y + c2*d_t, 1)
    where
    *d_y and d_t are residuals from the treatment and outcome model
    *c1 and c2 are coefficients to the residuals

    3. The final U, which is the simulated unobserved confounder is obtained by debiasing the intermediate variable U by residualising it with X


    Choosing the coefficients c1 and c2:
    The coefficients are chosen based on these basic assumptions:
    1. There is a hyperbolic relationship satisfying c1*c2 = c_star
    2. c_star is chosen from a range of possible values based on the correlation of the obtained simulated variable with outcome and treatment.
    3. The product of correlations with treatment and outcome should be at a minimum distance to the maximum correlations with treatment and outcome in any of the observed confounders
    4. The ratio of the weights should be such that they maintain the ratio of the maximum possible observed coefficients within some confidence interval

    :param c_star_max: The maximum possible value for the hyperbolic curve on which the coefficients to the residuals lie. It defaults to 1000 in the code if not specified by the user.
        :type int
    :param convergence_threshold: The threshold to check the plateauing of the correlation while selecting a c_star. It defaults to 0.1 in the code if not specified by the user
        :type float

    :returns: The simulated values of the unobserved confounder based on the data
        :type pandas.core.series.Series

    """

    # Obtaining the list of observed variables
    required_variables = True
    observed_variables = choose_variables(required_variables, variables_of_interest)

    observed_variables_with_treatment_and_outcome = observed_variables + treatment_name + outcome_name

    # Taking a subset of the dataframe that has only observed variables
    data = data[observed_variables_with_treatment_and_outcome]

    # Residuals from the outcome model obtained by fitting a linear model
    y = data[outcome_name[0]]
    observed_variables_with_treatment = observed_variables + treatment_name
    X = data[observed_variables_with_treatment]
    model = sm.OLS(y, X.astype("float"))
    results = model.fit()
    residuals_y = y - results.fittedvalues
    d_y = list(pd.Series(residuals_y))

    # Residuals from the treatment model obtained by fitting a linear model
    t = data[treatment_name[0]].astype("int64")
    X = data[observed_variables]
    model = sm.OLS(t, X)
    results = model.fit()
    residuals_t = t - results.fittedvalues
    d_t = list(pd.Series(residuals_t))

    # Initialising product_cor_metric_observed with a really low value as finding maximum
    product_cor_metric_observed = -10000000000

    for i in observed_variables:
        current_obs_confounder = data[i]
        outcome_values = data[outcome_name[0]]
        correlation_y = current_obs_confounder.corr(outcome_values)
        treatment_values = t
        correlation_t = current_obs_confounder.corr(treatment_values)
        product_cor_metric_current = correlation_y * correlation_t
        if product_cor_metric_current >= product_cor_metric_observed:
            product_cor_metric_observed = product_cor_metric_current
            correlation_t_observed = correlation_t
            correlation_y_observed = correlation_y

    # The user has an option to give the the effect_strength_on_y and effect_strength_on_t which can be then used instead of maximum correlation with treatment and outcome in the observed variables as it specifies the desired effect.
    if kappa_t is not None:
        correlation_t_observed = kappa_t
    if kappa_y is not None:
        correlation_y_observed = kappa_y

    # Choosing a c_star based on the data.
    # The correlations stop increasing upon increasing c_star after a certain value, that is it plateaus and we choose the value of c_star to be the value it plateaus.

    correlation_y_list = []
    correlation_t_list = []
    product_cor_metric_simulated_list = []
    x_list = []

    step = int(c_star_max / 10)
    for i in range(0, int(c_star_max), step):
        c1 = math.sqrt(i)
        c2 = c1
        final_U = _generate_confounder_from_residuals(c1, c2, d_y, d_t, X)
        current_simulated_confounder = final_U
        outcome_values = data[outcome_name[0]]
        correlation_y = current_simulated_confounder.corr(outcome_values)
        correlation_y_list.append(correlation_y)

        treatment_values = t
        correlation_t = current_simulated_confounder.corr(treatment_values)
        correlation_t_list.append(correlation_t)

        product_cor_metric_simulated = correlation_y * correlation_t
        product_cor_metric_simulated_list.append(product_cor_metric_simulated)

        x_list.append(i)

    index = 1
    while index < len(correlation_y_list):
        if (correlation_y_list[index] - correlation_y_list[index - 1]) <= convergence_threshold:
            c_star = x_list[index]
            break
        index = index + 1

    # Choosing c1 and c2 based on the hyperbolic relationship once c_star is chosen by going over various combinations of c1 and c2 values and choosing the combination which
    # which maintains the minimum distance between the product of correlations of the simulated variable and the product of maximum correlations of one of the observed variables
    # and additionally checks if the ratio of the weights are such that they maintain the ratio of the maximum possible observed coefficients within some confidence interval

    # c1_final and c2_final are initialised to the values on the hyperbolic curve such that c1_final = c2_final  and c1_final*c2_final = c_star
    c1_final = math.sqrt(c_star)
    c2_final = math.sqrt(c_star)

    # initialising min_distance_between_product_cor_metrics to be a value greater than 1
    min_distance_between_product_cor_metrics = 1.5
    i = 0.05

    threshold = c_star / 0.05

    while i <= threshold:
        c2 = i
        c1 = c_star / c2
        final_U = _generate_confounder_from_residuals(c1, c2, d_y, d_t, X)

        current_simulated_confounder = final_U
        outcome_values = data[outcome_name[0]]
        correlation_y = current_simulated_confounder.corr(outcome_values)

        treatment_values = t
        correlation_t = current_simulated_confounder.corr(treatment_values)

        product_cor_metric_simulated = correlation_y * correlation_t

        if min_distance_between_product_cor_metrics >= abs(product_cor_metric_simulated - product_cor_metric_observed):
            min_distance_between_product_cor_metrics = abs(product_cor_metric_simulated - product_cor_metric_observed)
            additional_condition = correlation_y_observed / correlation_t_observed
            if ((c1 / c2) <= (additional_condition + 0.3 * additional_condition)) and (
                (c1 / c2) >= (additional_condition - 0.3 * additional_condition)
            ):  # choose minimum positive value
                c1_final = c1
                c2_final = c2

        i = i * 1.5

    """#closed form solution

    print("c_star_max before closed form", c_star_max)

    if max_correlation_with_t == -1000:
        c2 = 0
        c1 = c_star_max
    else:
        additional_condition = abs(max_correlation_with_y/max_correlation_with_t)
        print("additional_condition", additional_condition)
        c2 = math.sqrt(c_star_max/additional_condition)
        c1 = c_star_max/c2"""

    final_U = _generate_confounder_from_residuals(c1_final, c2_final, d_y, d_t, X)

    return final_U


def _generate_confounder_from_residuals(c1, c2, d_y, d_t, X):
    """
    This function takes the residuals from the treatment and outcome model and their coefficients and simulates the intermediate random variable U by taking
    the row wise normal distribution corresponding to each residual value and then debiasing the intermediate variable to get the final variable.

    :param c1: coefficient to the residual from the outcome model
    :type float
    :param c2: coefficient to the residual from the treatment model
    :type float
    :param d_y: residuals from the outcome model
    :type list
    :param d_t: residuals from the treatment model
    :type list

    :returns: The simulated values of the unobserved confounder based on the data
    :type pandas.core.series.Series

    """
    U = []

    for j in range(len(d_t)):
        simulated_variable_mean = c1 * d_y[j] + c2 * d_t[j]
        simulated_variable_stddev = 1
        U.append(np.random.normal(simulated_variable_mean, simulated_variable_stddev, 1))

    U = np.array(U)
    model = sm.OLS(U, X)
    results = model.fit()
    U = U.reshape(
        -1,
    )
    final_U = U - results.fittedvalues.values
    final_U = pd.Series(U)

    return final_U


def sensitivity_linear_partial_r2(
    data: pd.DataFrame,
    estimate: CausalEstimate,
    treatment_name: str,
    frac_strength_treatment: float = 1.0,
    frac_strength_outcome: float = 1.0,
    percent_change_estimate: float = 1.0,
    benchmark_common_causes: Optional[List[str]] = None,
    significance_level: Optional[float] = None,
    null_hypothesis_effect: Optional[float] = None,
    plot_estimate: bool = True,
) -> LinearSensitivityAnalyzer:
    """Add an unobserved confounder for refutation using Linear partial R2 methond (Sensitivity Analysis for linear models).

    :param data: pd.DataFrame: Data to run the refutation
    :param estimate: CausalEstimate: Estimate to run the refutation
    :param treatment_name: str: Name of the treatment
    :param frac_strength_treatment: float: This parameter decides the effect strength of the simulated confounder as a fraction of the effect strength of observed confounders on treatment. Defaults to 1.
    :param frac_strength_outcome: float: This parameter decides the effect strength of the simulated confounder as a fraction of the effect strength of observed confounders on outcome. Defaults to 1.
    :param percent_change_estimate: It is the percentage of reduction of treatment estimate that could alter the results (default = 1).
                                    if percent_change_estimate = 1, the robustness value describes the strength of association of confounders with treatment and outcome in order to reduce the estimate by 100% i.e bring it down to 0. (relevant only for Linear Sensitivity Analysis, ignore for rest)
    :param benchmark_common_causes: names of variables for bounding strength of confounders. (relevant only for partial-r2 based simulation methods)
    :param significance_level: confidence interval for statistical inference(default = 0.05). (relevant only for partial-r2 based simulation methods)
    :param null_hypothesis_effect: assumed effect under the null hypothesis. (relevant only for linear-partial-R2, ignore for rest)
    :param plot_estimate: Generate contour plot for estimate while performing sensitivity analysis. (default = True).
                            (relevant only for partial-r2 based simulation methods)

    """

    if not (isinstance(estimate.estimator, LinearRegressionEstimator)):
        raise NotImplementedError("Currently only LinearRegressionEstimator is supported for Sensitivity Analysis")

    if estimate.estimator._effect_modifier_names is not None and len(estimate.estimator._effect_modifier_names) > 0:
        raise NotImplementedError("The current implementation does not support effect modifiers")

    if frac_strength_outcome == 1:
        frac_strength_outcome = frac_strength_treatment

    analyzer = LinearSensitivityAnalyzer(
        estimator=estimate.estimator,
        data=data,
        treatment_name=treatment_name,
        percent_change_estimate=percent_change_estimate,
        significance_level=significance_level,
        benchmark_common_causes=benchmark_common_causes,
        null_hypothesis_effect=null_hypothesis_effect,
        frac_strength_treatment=frac_strength_treatment,
        frac_strength_outcome=frac_strength_outcome,
        common_causes_order=estimate.estimator._observed_common_causes.columns,
    )

    analyzer.check_sensitivity(plot=plot_estimate)
    return analyzer


def sensitivity_non_parametric_partial_r2(
    estimate: CausalEstimate,
    kappa_t: Optional[Union[float, np.ndarray]] = None,
    kappa_y: Optional[Union[float, np.ndarray]] = None,
    frac_strength_treatment: float = 1.0,
    frac_strength_outcome: float = 1.0,
    benchmark_common_causes: Optional[List[str]] = None,
    plot_estimate: bool = True,
    alpha_s_estimator_list: Optional[List] = None,
    alpha_s_estimator_param_list: Optional[List[Dict]] = None,
    g_s_estimator_list: Optional[List] = None,
    g_s_estimator_param_list: Optional[List[Dict]] = None,
    plugin_reisz: bool = False,
) -> Union[PartialLinearSensitivityAnalyzer, NonParametricSensitivityAnalyzer]:
    """Add an unobserved confounder for refutation using Non-parametric partial R2 methond (Sensitivity Analysis for non-parametric models).

    :param estimate: CausalEstimate: Estimate to run the refutation
    :param kappa_t: float, numpy.ndarray: Partial R2 of the unobserved confounder wrt the treatment conditioned on the observed confounders. Only in the case of general non-parametric-partial-R2, it is the fraction of variance in the reisz representer that is explained by the unobserved confounder; specifically  (1-r), where r is the ratio of variance of reisz representer, alpha^2, based on observed confounders and that based on all confounders.
    :param kappa_y: float, numpy.ndarray: Partial R2 of the unobserved confounder wrt the outcome conditioned on the treatment and observed confounders.
    :param frac_strength_treatment: float: This parameter decides the effect strength of the simulated confounder as a fraction of the effect strength of observed confounders on treatment. Defaults to 1.
    :param frac_strength_outcome: float: This parameter decides the effect strength of the simulated confounder as a fraction of the effect strength of observed confounders on outcome. Defaults to 1.
    :param benchmark_common_causes: names of variables for bounding strength of confounders. (relevant only for partial-r2 based simulation methods)
    :param plot_estimate: Generate contour plot for estimate while performing sensitivity analysis. (default = True).
                            (relevant only for partial-r2 based simulation methods)
    :param alpha_s_estimator_list: list of estimator objects for estimating alpha_s. These objects should have fit() and predict() methods (relevant only for non-parametric-partial-R2 method)
    :param alpha_s_estimator_param_list: list of dictionaries with parameters for finding alpha_s. (relevant only for non-parametric-partial-R2 simulation method)
    :param g_s_estimator_list: list of estimator objects for finding g_s. These objects should have fit() and predict() functions implemented. (relevant only for non-parametric-partial-R2 simulation method)
    :param g_s_estimator_param_list: list of dictionaries with parameters for tuning respective estimators in "g_s_estimator_list". The order of the dictionaries in the list should be consistent with the estimator objects order in "g_s_estimator_list". (relevant only for non-parametric-partial-R2 simulation method)
    :plugin_reisz: bool: Flag on whether to use the plugin estimator or the nonparametric estimator for reisz representer function (alpha_s).
    """

    import dowhy.causal_estimators.econml

    # If the estimator used is LinearDML, partially linear sensitivity analysis will be automatically chosen
    if (
        isinstance(estimate.estimator, dowhy.causal_estimators.econml.Econml)
        and estimate.estimator.estimator.__class__.__name__ == "LinearDML"
    ):

        analyzer = PartialLinearSensitivityAnalyzer(
            estimator=estimate._estimator_object,
            observed_common_causes=estimate.estimator._observed_common_causes,
            treatment=estimate.estimator._treatment,
            outcome=estimate.estimator._outcome,
            alpha_s_estimator_param_list=alpha_s_estimator_param_list,
            g_s_estimator_list=g_s_estimator_list,
            g_s_estimator_param_list=g_s_estimator_param_list,
            effect_strength_treatment=kappa_t,
            effect_strength_outcome=kappa_y,
            benchmark_common_causes=benchmark_common_causes,
            frac_strength_treatment=frac_strength_treatment,
            frac_strength_outcome=frac_strength_outcome,
        )
        analyzer.check_sensitivity(plot=plot_estimate)
        return analyzer

    analyzer = NonParametricSensitivityAnalyzer(
        estimator=estimate.estimator,
        observed_common_causes=estimate.estimator._observed_common_causes,
        treatment=estimate.estimator._treatment,
        outcome=estimate.estimator._outcome,
        alpha_s_estimator_list=alpha_s_estimator_list,
        alpha_s_estimator_param_list=alpha_s_estimator_param_list,
        g_s_estimator_list=g_s_estimator_list,
        g_s_estimator_param_list=g_s_estimator_param_list,
        effect_strength_treatment=kappa_t,
        effect_strength_outcome=kappa_y,
        benchmark_common_causes=benchmark_common_causes,
        frac_strength_treatment=frac_strength_treatment,
        frac_strength_outcome=frac_strength_outcome,
        theta_s=estimate.value,
        plugin_reisz=plugin_reisz,
    )
    analyzer.check_sensitivity(plot=plot_estimate)
    return analyzer


def sensitivity_e_value(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    estimate: CausalEstimate,
    treatment_name: List[str],
    outcome_name: List[str],
    plot_estimate: bool = True,
) -> EValueSensitivityAnalyzer:
    if not isinstance(estimate.estimator, RegressionEstimator):
        raise NotImplementedError("E-Value sensitivity analysis is currently only implemented RegressionEstimator.")

    if estimate.estimator._effect_modifier_names is not None and len(estimate.estimator._effect_modifier_names) > 0:
        raise NotImplementedError("The current implementation does not support effect modifiers")

    analyzer = EValueSensitivityAnalyzer(
        estimate=estimate,
        estimand=target_estimand,
        data=data,
        treatment_name=treatment_name[0],
        outcome_name=outcome_name[0],
    )
    analyzer.check_sensitivity(plot=plot_estimate)
    return analyzer


def sensitivity_simulation(
    data: pd.DataFrame,
    target_estimand: IdentifiedEstimand,
    estimate: CausalEstimate,
    treatment_name: str,
    outcome_name: str,
    kappa_t: Optional[Union[float, np.ndarray]] = None,
    kappa_y: Optional[Union[float, np.ndarray]] = None,
    confounders_effect_on_treatment: str = "binary_flip",
    confounders_effect_on_outcome: str = "linear",
    frac_strength_treatment: float = 1.0,
    frac_strength_outcome: float = 1.0,
    plotmethod: Optional[str] = None,
    show_progress_bar=False,
    **_,
) -> CausalRefutation:
    """
    This function attempts to add an unobserved common cause to the outcome and the treatment. At present, we have implemented the behavior for one dimensional behaviors for continuous
    and binary variables. This function can either take single valued inputs or a range of inputs. The function then looks at the data type of the input and then decides on the course of
    action.

    :param data: pd.DataFrame: Data to run the refutation
    :param target_estimand: IdentifiedEstimand: Identified estimand to run the refutation
    :param estimate: CausalEstimate: Estimate to run the refutation
    :param treatment_name: str: Name of the treatment
    :param outcome_name: str: Name of the outcome
    :param kappa_t: float, numpy.ndarray: Strength of the confounder's effect on treatment. When confounders_effect_on_treatment is linear,  it is the regression coefficient. When the confounders_effect_on_treatment is binary flip, it is the probability with which effect of unobserved confounder can invert the value of the treatment.
    :param kappa_y: float, numpy.ndarray: Strength of the confounder's effect on outcome. Its interpretation depends on confounders_effect_on_outcome and the simulation_method. When simulation_method is direct-simulation, for a linear effect it behaves like the regression coefficient and for a binary flip, it is the probability with which it can invert the value of the outcome.
    :param confounders_effect_on_treatment: str : The type of effect on the treatment due to the unobserved confounder. Possible values are ['binary_flip', 'linear']
    :param confounders_effect_on_outcome: str : The type of effect on the outcome due to the unobserved confounder. Possible values are ['binary_flip', 'linear']
    :param frac_strength_treatment: float: This parameter decides the effect strength of the simulated confounder as a fraction of the effect strength of observed confounders on treatment. Defaults to 1.
    :param frac_strength_outcome: float: This parameter decides the effect strength of the simulated confounder as a fraction of the effect strength of observed confounders on outcome. Defaults to 1.
    :param plotmethod: string: Type of plot to be shown. If None, no plot is generated. This parameter is used only only when more than one treatment confounder effect values or outcome confounder effect values are provided. Default is "colormesh". Supported values are "contour", "colormesh" when more than one value is provided for both confounder effect value parameters; "line" when provided for only one of them.

    :return: CausalRefuter: An object that contains the estimated effect and a new effect and the name of the refutation used.
    """
    if kappa_t is None:
        kappa_t = _infer_default_kappa_t(
            data, target_estimand, treatment_name, confounders_effect_on_treatment, frac_strength_treatment
        )
    if kappa_y is None:
        kappa_y = _infer_default_kappa_y(
            data, target_estimand, outcome_name, confounders_effect_on_outcome, frac_strength_outcome
        )

    if not isinstance(kappa_t, (list, np.ndarray)) and not isinstance(
        kappa_y, (list, np.ndarray)
    ):  # Deal with single value inputs
        new_data = copy.deepcopy(data)
        new_data = _include_confounders_effect(
            data,
            new_data,
            confounders_effect_on_treatment,
            treatment_name,
            kappa_t,
            confounders_effect_on_outcome,
            outcome_name,
            kappa_y,
        )
        new_estimator = estimate.estimator.get_new_estimator_object(target_estimand)
        new_estimator.fit(
            new_data,
            target_estimand.treatment_variable,
            target_estimand.outcome_variable,
            estimate.estimator._effect_modifier_names,
            **new_estimator._econml_fit_params if isinstance(new_estimator, Econml) else {},
        )
        new_effect = new_estimator.estimate_effect(
            control_value=estimate.control_value,
            treatment_value=estimate.treatment_value,
            target_units=estimate.estimator._target_units,
        )
        refute = CausalRefutation(
            estimate.value, new_effect.value, refutation_type="Refute: Add an Unobserved Common Cause"
        )

        refute.new_effect_array = np.array(new_effect.value)
        refute.new_effect = new_effect.value
        return refute

    else:  # Deal with multiple value inputs

        if isinstance(kappa_t, (list, np.ndarray)) and isinstance(
            kappa_y, (list, np.ndarray)
        ):  # Deal with range inputs
            # Get a 2D matrix of values
            # x,y =  np.meshgrid(self.kappa_t, self.kappa_y) # x,y are both MxN

            results_matrix = np.random.rand(len(kappa_t), len(kappa_y))  # Matrix to hold all the results of NxM
            orig_data = copy.deepcopy(data)

            for i in tqdm(
                range(len(kappa_t)),
                colour=CausalRefuter.PROGRESS_BAR_COLOR,
                disable=not show_progress_bar,
                desc="Refuting Estimates: ",
            ):
                for j in range(len(kappa_y)):
                    new_data = _include_confounders_effect(
                        data,
                        orig_data,
                        confounders_effect_on_treatment,
                        treatment_name,
                        kappa_t[i],
                        confounders_effect_on_outcome,
                        outcome_name,
                        kappa_y[j],
                    )
                    new_estimator = estimate.estimator.get_new_estimator_object(target_estimand)
                    new_estimator.fit(
                        new_data,
                        target_estimand.treatment_variable,
                        target_estimand.outcome_variable,
                        estimate.estimator._effect_modifier_names,
                        **new_estimator._econml_fit_params if isinstance(new_estimator, Econml) else {},
                    )
                    new_effect = new_estimator.estimate_effect(
                        control_value=estimate.control_value,
                        treatment_value=estimate.treatment_value,
                        target_units=estimate.estimator._target_units,
                    )
                    refute = CausalRefutation(
                        estimate.value,
                        new_effect.value,
                        refutation_type="Refute: Add an Unobserved Common Cause",
                    )
                    results_matrix[i][j] = refute.new_effect  # Populate the results

            refute.new_effect_array = results_matrix
            refute.new_effect = (np.min(results_matrix), np.max(results_matrix))
            # Store the values into the refute object
            if plotmethod is None:
                return refute

            import matplotlib
            import matplotlib.pyplot as plt

            fig = plt.figure(figsize=(6, 5))
            left, bottom, width, height = 0.1, 0.1, 0.8, 0.8
            ax = fig.add_axes([left, bottom, width, height])

            oe = estimate.value
            contour_levels = [oe / 4.0, oe / 2.0, (3.0 / 4) * oe, oe]
            contour_levels.extend([0, np.min(results_matrix), np.max(results_matrix)])
            if plotmethod == "contour":
                cp = plt.contourf(kappa_y, kappa_t, results_matrix, levels=sorted(contour_levels))
                # Adding a label on the contour line for the original estimate
                fmt = {}
                trueeffect_index = np.where(cp.levels == oe)[0][0]
                fmt[cp.levels[trueeffect_index]] = "Estimated Effect"
                # Label every other level using strings
                plt.clabel(cp, [cp.levels[trueeffect_index]], inline=True, fmt=fmt)
                plt.colorbar(cp)
            elif plotmethod == "colormesh":
                cp = plt.pcolormesh(kappa_y, kappa_t, results_matrix, shading="nearest")
                plt.colorbar(cp, ticks=contour_levels)
            ax.yaxis.set_ticks(kappa_t)
            ax.xaxis.set_ticks(kappa_y)
            plt.xticks(rotation=45)
            ax.set_title("Effect of Unobserved Common Cause")
            ax.set_ylabel("Value of Linear Constant on Treatment")
            ax.set_xlabel("Value of Linear Constant on Outcome")
            plt.show()

            return refute

        elif isinstance(kappa_t, (list, np.ndarray)):
            outcomes = np.random.rand(len(kappa_t))
            orig_data = copy.deepcopy(data)

            for i in tqdm(
                range(0, len(kappa_t)),
                colour=CausalRefuter.PROGRESS_BAR_COLOR,
                disable=not show_progress_bar,
                desc="Refuting Estimates: ",
            ):
                new_data = _include_confounders_effect(
                    data,
                    orig_data,
                    confounders_effect_on_treatment,
                    treatment_name,
                    kappa_t[i],
                    confounders_effect_on_outcome,
                    outcome_name,
                    kappa_y,
                )
                new_estimator = estimate.estimator.get_new_estimator_object(target_estimand)
                new_estimator.fit(
                    new_data,
                    target_estimand.treatment_variable,
                    target_estimand.outcome_variable,
                    estimate.estimator._effect_modifier_names,
                    **new_estimator._econml_fit_params if isinstance(new_estimator, Econml) else {},
                )
                new_effect = new_estimator.estimate_effect(
                    control_value=estimate.control_value,
                    treatment_value=estimate.treatment_value,
                    target_units=estimate.estimator._target_units,
                )
                refute = CausalRefutation(
                    estimate.value, new_effect.value, refutation_type="Refute: Add an Unobserved Common Cause"
                )
                logger.debug(refute)
                outcomes[i] = refute.new_effect  # Populate the results

            refute.new_effect_array = outcomes
            refute.new_effect = (np.min(outcomes), np.max(outcomes))
            if plotmethod is None:
                return refute

            import matplotlib
            import matplotlib.pyplot as plt

            fig = plt.figure(figsize=(6, 5))
            left, bottom, width, height = 0.1, 0.1, 0.8, 0.8
            ax = fig.add_axes([left, bottom, width, height])

            plt.plot(kappa_t, outcomes)
            plt.axhline(estimate.value, linestyle="--", color="gray")
            ax.set_title("Effect of Unobserved Common Cause")
            ax.set_xlabel("Value of Linear Constant on Treatment")
            ax.set_ylabel("Estimated Effect after adding the common cause")
            plt.show()

            return refute

        elif isinstance(kappa_y, (list, np.ndarray)):
            outcomes = np.random.rand(len(kappa_y))
            orig_data = copy.deepcopy(data)

            for i in tqdm(
                range(0, len(kappa_y)),
                colour=CausalRefuter.PROGRESS_BAR_COLOR,
                disable=not show_progress_bar,
                desc="Refuting Estimates: ",
            ):
                new_data = _include_confounders_effect(
                    data,
                    orig_data,
                    confounders_effect_on_treatment,
                    treatment_name,
                    kappa_t,
                    confounders_effect_on_outcome,
                    outcome_name,
                    kappa_y[i],
                )
                new_estimator = estimate.estimator.get_new_estimator_object(target_estimand)
                new_estimator.fit(
                    new_data,
                    target_estimand.treatment_variable,
                    target_estimand.outcome_variable,
                    estimate.estimator._effect_modifier_names,
                    **new_estimator._econml_fit_params if isinstance(new_estimator, Econml) else {},
                )
                new_effect = new_estimator.estimate_effect(
                    control_value=estimate.control_value,
                    treatment_value=estimate.treatment_value,
                    target_units=estimate.estimator._target_units,
                )
                refute = CausalRefutation(
                    estimate.value, new_effect.value, refutation_type="Refute: Add an Unobserved Common Cause"
                )
                logger.debug(refute)
                outcomes[i] = refute.new_effect  # Populate the results

            refute.new_effect_array = outcomes
            refute.new_effect = (np.min(outcomes), np.max(outcomes))
            if plotmethod is None:
                return refute

            import matplotlib
            import matplotlib.pyplot as plt

            fig = plt.figure(figsize=(6, 5))
            left, bottom, width, height = 0.1, 0.1, 0.8, 0.8
            ax = fig.add_axes([left, bottom, width, height])

            plt.plot(kappa_y, outcomes)
            plt.axhline(estimate.value, linestyle="--", color="gray")
            ax.set_title("Effect of Unobserved Common Cause")
            ax.set_xlabel("Value of Linear Constant on Outcome")
            ax.set_ylabel("Estimated Effect after adding the common cause")
            plt.show()

            return refute



"""Functions to estimate uncertainties such as entropy, KL divergence etc.

Functions in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""

import numpy as np
from numpy.dual import det
from scipy.special import digamma
from scipy.stats import entropy
from sklearn.neighbors import NearestNeighbors

from dowhy.gcm.constant import EPS
from dowhy.gcm.util.general import shape_into_2d


def estimate_entropy_using_discretization(X: np.ndarray, bin_width: float = 1) -> float:
    # TODO: Add method for auto select a bin_width/width based on the data. Make sure that the auto selection method is
    #  theoretically sound, i.e. make entropy results from different data comparable.
    X = shape_into_2d(X)

    if X.shape[1] > 1:
        raise RuntimeError(
            "The discrete entropy estimator can only handle one dimensional data, but the input data is "
            "%d dimensional!" % X.shape[1]
        )

    max_value = np.max(X)
    min_value = np.min(X)
    number_of_bins = max(1, int((max_value - min_value) / bin_width))
    num_samples = X.shape[0]
    return -np.sum(
        [
            (i / num_samples * np.log(i / num_samples)) if i > 0 else 0
            for i in np.histogram(X, bins=np.linspace(min_value, max_value, number_of_bins).reshape(-1))[0]
        ]
    )


def estimate_entropy_kmeans(X: np.ndarray) -> float:
    """Related paper:
    Kozachenko, L., & Leonenko, N. (1987). Sample estimate of the entropy of a random vector. Problemy Peredachi
    Informatsii, 23(2), 9–16.
    """
    X = shape_into_2d(X)

    k = int(np.sqrt(X.shape[0]))

    x_neighbourhood = NearestNeighbors(n_neighbors=k + 1).fit(X)
    distances, _ = x_neighbourhood.kneighbors(X, k + 1)
    distances = distances[:, -1]

    sum_log_dist = np.sum(np.log(2 * distances + EPS))

    return -digamma(k) + digamma(X.shape[0]) + (X.shape[1] / float(X.shape[0])) * sum_log_dist


def estimate_gaussian_entropy(X: np.ndarray) -> float:
    """Entropy with respect to standardized variables."""
    X = shape_into_2d(X)

    if X.shape[1] > 1:
        return 0.5 * np.log((2 * np.pi * np.e) ** X.shape[1] * estimate_variance(X))
    else:
        return 0.5 * np.log(2 * np.pi * np.e * estimate_variance(X))


def estimate_variance(X: np.ndarray) -> float:
    X = shape_into_2d(X)

    if X.shape[1] > 1:
        result = det(np.cov(X, rowvar=False))
    else:
        result = np.var(X)

    # Extremely small values can somehow result in negative values.
    return max(0.0, result)


def estimate_entropy_of_probabilities(X: np.ndarray) -> float:
    """Estimates the entropy of each probability vector (row wise) in X separately and returns the mean over all
    results.
    """
    return float(np.mean(entropy(X, axis=1)))


def estimate_entropy_discrete(X: np.ndarray) -> float:
    """Estimates the entropy assuming the data in X is discrete.

    :param X: Discrete samples.
    :return: Entropy of X.
    """
    X = shape_into_2d(X)

    _, counts = np.unique(X, return_counts=True, axis=0)
    return entropy(counts)



import numpy as np
import pandas as pd
import pytest
from flaky import flaky
from sklearn.ensemble import RandomForestRegressor as RFR
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LinearRegression

from dowhy.gcm.ml.regression import SklearnRegressionModel
from dowhy.gcm.unit_change import (
    SklearnLinearRegressionModel,
    unit_change,
    unit_change_linear,
    unit_change_linear_input_only,
    unit_change_nonlinear,
    unit_change_nonlinear_input_only,
)


@flaky(max_runs=5)
def test_given_fitted_linear_mechanisms_with_output_change_when_evaluate_unit_change_linear_method_then_returns_correct_attributions():
    num_rows = 100
    A1 = np.random.normal(size=num_rows)
    B1 = np.random.normal(size=num_rows)
    C1 = 2 * A1 + 3 * B1

    A2 = np.random.normal(size=num_rows)
    B2 = np.random.normal(size=num_rows)
    C2 = 3 * A2 + 2 * B2

    background_df = pd.DataFrame(data=dict(A=A1, B=B1))
    foreground_df = pd.DataFrame(data=dict(A=A2, B=B2))

    background_mechanism = SklearnLinearRegressionModel(
        LinearRegression(fit_intercept=False).fit(np.column_stack((A1, B1)), C1)
    )
    foreground_mechanism = SklearnLinearRegressionModel(
        LinearRegression(fit_intercept=False).fit(np.column_stack((A2, B2)), C2)
    )

    actual_contributions = unit_change_linear(
        background_mechanism, background_df, foreground_mechanism, foreground_df, ["A", "B"]
    )
    expected_contributions = pd.DataFrame(
        data=dict(
            A=(3 + 2) * (A2 - A1) / 2, B=(2 + 3) * (B2 - B1) / 2, f=(A1 + A2) * (3 - 2) / 2 + (B1 + B2) * (2 - 3) / 2
        )
    )

    np.testing.assert_array_almost_equal(actual_contributions, expected_contributions, decimal=1)


@flaky(max_runs=5)
def test_given_fitted_linear_mechanisms_with_output_change_when_evaluate_unit_change_linear_and_nonlinear_methods_then_attributions_are_consistent():
    num_rows = 100
    A1 = np.random.normal(size=num_rows)
    B1 = np.random.normal(size=num_rows)
    C1 = 2 * A1 + 3 * B1

    A2 = np.random.normal(size=num_rows)
    B2 = np.random.normal(size=num_rows)
    C2 = 3 * A2 + 2 * B2

    background_df = pd.DataFrame(data=dict(A=A1, B=B1))
    foreground_df = pd.DataFrame(data=dict(A=A2, B=B2))

    background_mechanism = SklearnLinearRegressionModel(
        LinearRegression(fit_intercept=False).fit(np.column_stack((A1, B1)), C1)
    )
    foreground_mechanism = SklearnLinearRegressionModel(
        LinearRegression(fit_intercept=False).fit(np.column_stack((A2, B2)), C2)
    )

    actual_contributions = unit_change_linear(
        background_mechanism, background_df, foreground_mechanism, foreground_df, ["A", "B"]
    )
    expected_contributions = unit_change_nonlinear(
        background_mechanism, background_df, foreground_mechanism, foreground_df, ["A", "B"]
    )

    np.testing.assert_array_almost_equal(actual_contributions, expected_contributions, decimal=1)


def test_given_unfitted_mechanisms_when_evaluate_unit_change_methods_then_raises_exception():
    with pytest.raises(NotFittedError):
        unit_change_linear(
            SklearnLinearRegressionModel(LinearRegression()),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            SklearnLinearRegressionModel(LinearRegression()),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            ["A", "B"],
        )

    with pytest.raises(NotFittedError):
        unit_change_nonlinear(
            SklearnRegressionModel(LinearRegression()),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            SklearnRegressionModel(LinearRegression()),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            ["A", "B"],
        )

    with pytest.raises(NotFittedError):
        unit_change_nonlinear(
            SklearnRegressionModel(RFR()),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            SklearnRegressionModel(RFR()),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            ["A", "B"],
        )


def test_given_fitted_nonlinnear_mechanisms_when_evaluate_unit_change_linear_method_then_raises_exception():
    with pytest.raises(AttributeError):
        unit_change_linear(
            SklearnRegressionModel(RFR().fit(np.random.normal(size=(100, 2)), np.random.normal(size=100))),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            SklearnRegressionModel(RFR().fit(np.random.normal(size=(100, 2)), np.random.normal(size=100))),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            ["A", "B"],
        )


@flaky(max_runs=5)
def test_given_fitted_mechanisms_with_no_input_change_when_evaluate_unit_change_input_only_methods_then_returns_zero_attributions():
    num_rows = 100
    A = np.random.normal(size=num_rows)
    B = np.random.normal(size=num_rows)
    C = 3 * A + 2 * B

    background_df = pd.DataFrame(data=dict(A=A, B=B, C=C))
    foreground_df = pd.DataFrame(data=dict(A=A, B=B, C=C))

    actual_contributions = unit_change_nonlinear_input_only(
        SklearnRegressionModel(RFR().fit(np.column_stack((A, B)), C)),
        background_df,
        foreground_df,
        ["A", "B"],
    )
    expected_contributions = pd.DataFrame(data=dict(A=np.zeros(num_rows), B=np.zeros(num_rows)))
    np.testing.assert_array_almost_equal(actual_contributions, expected_contributions, decimal=1)


@flaky(max_runs=5)
def test_given_fitted_linear_mechanism_with_input_change_when_evaluate_unit_change_linear_input_only_then_returns_correct_attributions():
    num_rows = 100
    A1 = np.random.normal(size=num_rows)
    B1 = np.random.normal(size=num_rows)
    C1 = 3 * A1 + 2 * B1

    A2 = np.random.normal(size=num_rows)
    B2 = np.random.normal(size=num_rows)
    C2 = 3 * A2 + 2 * B2

    background_df = pd.DataFrame(data=dict(A=A1, B=B1, C=C1))
    foreground_df = pd.DataFrame(data=dict(A=A2, B=B2, C=C2))

    fitted_linear_reg = LinearRegression()
    fitted_linear_reg.coef_ = np.array([3, 2])

    actual_contributions = unit_change_linear_input_only(
        SklearnLinearRegressionModel(fitted_linear_reg), background_df, foreground_df, ["A", "B"]
    )
    expected_contributions = pd.DataFrame(data=dict(A=3 * (A2 - A1), B=2 * (B2 - B1)))
    np.testing.assert_array_almost_equal(actual_contributions, expected_contributions, decimal=1)


@flaky(max_runs=5)
def test_given_fitted_linear_mechanism_with_input_change_when_evaluate_unit_change_input_only_methods_then_attributions_are_consistent():
    num_rows = 100
    A1 = np.random.normal(size=num_rows)
    B1 = np.random.normal(size=num_rows)
    C1 = 3 * A1 + 2 * B1

    A2 = np.random.normal(size=num_rows)
    B2 = np.random.normal(size=num_rows)
    C2 = 3 * A2 + 2 * B2

    background_df = pd.DataFrame(data=dict(A=A1, B=B1, C=C1))
    foreground_df = pd.DataFrame(data=dict(A=A2, B=B2, C=C2))

    mechanism = SklearnLinearRegressionModel(LinearRegression().fit(np.column_stack((A1, B1)), C1))
    actual_contributions = unit_change_nonlinear_input_only(mechanism, background_df, foreground_df, ["A", "B"])
    expected_contributions = unit_change_linear_input_only(mechanism, background_df, foreground_df, ["A", "B"])

    np.testing.assert_array_almost_equal(actual_contributions, expected_contributions, decimal=1)


def test_given_unfitted_mechanisms_when_evaluate_unit_change_input_only_methods_then_raises_exception():
    with pytest.raises(NotFittedError):
        unit_change_linear_input_only(
            SklearnLinearRegressionModel(LinearRegression()),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            ["A", "B"],
        )

    with pytest.raises(NotFittedError):
        unit_change_nonlinear_input_only(
            SklearnLinearRegressionModel(LinearRegression()),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            ["A", "B"],
        )

    with pytest.raises(NotFittedError):
        unit_change_nonlinear_input_only(
            SklearnRegressionModel(RFR()),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            ["A", "B"],
        )


def test_given_fitted_nonlinnear_mechanism_when_evaluate_unit_change_linear_input_only_method_then_raises_exception():
    with pytest.raises(AttributeError):
        unit_change_linear_input_only(
            SklearnRegressionModel(RFR().fit(np.random.normal(size=(100, 2)), np.random.normal(size=100))),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            pd.DataFrame(data=dict(A=np.random.normal(size=100), B=np.random.normal(size=100))),
            ["A", "B"],
        )


@flaky(max_runs=5)
def test_given_single_mechanism_with_default_optional_parameters_when_evaluate_unit_change_then_returns_correct_attributions_to_input_only():
    num_rows = 100
    A1 = np.random.normal(size=num_rows)
    B1 = np.random.normal(size=num_rows)
    C1 = 2 * A1 + 3 * B1

    A2 = np.random.normal(size=num_rows)
    B2 = np.random.normal(size=num_rows)
    # C2 = 3 * A2 + 2 * B2

    background_df = pd.DataFrame(data=dict(A=A1, B=B1))
    foreground_df = pd.DataFrame(data=dict(A=A2, B=B2))

    mechanism = SklearnLinearRegressionModel(LinearRegression(fit_intercept=False).fit(np.column_stack((A1, B1)), C1))

    actual_contributions = unit_change(background_df, foreground_df, ["A", "B"], mechanism)
    expected_contributions = unit_change_linear_input_only(mechanism, background_df, foreground_df, ["A", "B"])

    np.testing.assert_array_almost_equal(actual_contributions, expected_contributions, decimal=1)

    mechanism = SklearnRegressionModel(RFR().fit(np.column_stack((A1, B1)), C1))

    actual_contributions = unit_change(background_df, foreground_df, ["A", "B"], mechanism)
    expected_contributions = unit_change_nonlinear_input_only(mechanism, background_df, foreground_df, ["A", "B"])
    np.testing.assert_array_almost_equal(actual_contributions, expected_contributions, decimal=1)


@flaky(max_runs=5)
def test_given_two_mechanisms_when_evaluate_unit_change_then_returns_correct_attributions_to_both_mechanism_and_input():
    num_rows = 100
    A1 = np.random.normal(size=num_rows)
    B1 = np.random.normal(size=num_rows)
    C1 = 2 * A1 + 3 * B1

    A2 = np.random.normal(size=num_rows)
    B2 = np.random.normal(size=num_rows)
    C2 = 3 * A2 + 2 * B2

    background_df = pd.DataFrame(data=dict(A=A1, B=B1))
    foreground_df = pd.DataFrame(data=dict(A=A2, B=B2))

    background_mechanism = SklearnLinearRegressionModel(
        LinearRegression(fit_intercept=False).fit(np.column_stack((A1, B1)), C1)
    )
    foreground_mechanism = SklearnLinearRegressionModel(
        LinearRegression(fit_intercept=False).fit(np.column_stack((A2, B2)), C2)
    )

    actual_contributions = unit_change(
        background_df, foreground_df, ["A", "B"], background_mechanism, foreground_mechanism
    )
    expected_contributions = unit_change_linear(
        background_mechanism, background_df, foreground_mechanism, foreground_df, ["A", "B"]
    )

    np.testing.assert_array_almost_equal(actual_contributions, expected_contributions, decimal=1)

    background_mechanism = SklearnRegressionModel(RFR().fit(np.column_stack((A1, B1)), C1))
    foreground_mechanism = SklearnRegressionModel(RFR().fit(np.column_stack((A2, B2)), C2))

    actual_contributions = unit_change(
        background_df, foreground_df, ["A", "B"], background_mechanism, foreground_mechanism
    )
    expected_contributions = unit_change_nonlinear(
        background_mechanism, background_df, foreground_mechanism, foreground_df, ["A", "B"]
    )

    np.testing.assert_array_almost_equal(actual_contributions, expected_contributions, decimal=1)



"""This module defines functions to attribute distribution changes.

Functions in this module should be considered experimental, meaning there might be breaking API changes in the future.
"""

import logging
from typing import Any, Callable, Dict, Optional, Tuple, Union

import networkx as nx
import numpy as np
import pandas as pd
from numpy.matlib import repmat
from statsmodels.stats.multitest import multipletests
from tqdm import tqdm

from dowhy.gcm.auto import AssignmentQuality, assign_causal_mechanisms
from dowhy.gcm.cms import ProbabilisticCausalModel
from dowhy.gcm.divergence import auto_estimate_kl_divergence
from dowhy.gcm.fitting_sampling import draw_samples, fit_causal_model_of_target
from dowhy.gcm.graph import (
    PARENTS_DURING_FIT,
    ConditionalStochasticModel,
    DirectedGraph,
    clone_causal_models,
    get_ordered_predecessors,
    is_root_node,
    node_connected_subgraph_view,
    validate_causal_dag,
)
from dowhy.gcm.independence_test.kernel import kernel_based
from dowhy.gcm.shapley import ShapleyConfig, estimate_shapley_values
from dowhy.gcm.util.general import shape_into_2d

_logger = logging.getLogger(__name__)


def mechanism_change_test(
    target_original_data: np.ndarray,
    target_new_data: np.ndarray,
    parents_original_data: Optional[np.ndarray] = None,
    parents_new_data: Optional[np.ndarray] = None,
    independence_test: Callable[[np.ndarray, np.ndarray], float] = kernel_based,
    conditional_independence_test: Callable[[np.ndarray, np.ndarray, np.ndarray], float] = kernel_based,
) -> float:
    """Estimates a p-value for the null hypothesis that the original and new data were generated by the same mechanism.
    Here, we check the dependency between binary labels indicating whether a sample is from the original or a new
    data set. If the labels do not provide information to determine if a sample is coming from the original/new
    distribution, then it is likely that the mechanism has not changed.

    For non-root nodes, samples from parent variables are needed as conditioning variables. This is, testing the
    null hypothesis that the data were generated by the same mechanism given the parent samples. By this, we incorporate
    upstream changes that might have impacted the parents, but not the target node itself.

    :param target_original_data: Samples of the node from the original data set.
    :param target_new_data: Samples of the node from the new data set.
    :param parents_original_data: Samples from parents of the node from the original data set.
    :param parents_new_data: Samples from parents of the node from the new data set.
    :param independence_test: Unconditional independence test. This is used to identify mechanism changes in nodes
                              without parents.
    :param conditional_independence_test: Conditional independence test. This is used to identify mechanism changes in
                                          nodes with parents.
    :return: A p-value for the null hypothesis that the mechanism has not changed.
    """
    causal_graph: DirectedGraph

    if parents_original_data is not None and parents_new_data is None:
        raise ValueError("Original parents data were given, but no new parents data!")
    if parents_original_data is None and parents_new_data is not None:
        raise ValueError("New parents data were given, but no original parents data!")

    num_samples_for_testing = min(target_original_data.shape[0], target_new_data.shape[0])
    data_set_indices = np.ones(num_samples_for_testing * 2)
    data_set_indices[num_samples_for_testing:] = -1
    data_set_indices = data_set_indices.astype(str)

    original_indices = np.random.choice(target_original_data.shape[0], num_samples_for_testing, replace=False)
    new_indices = np.random.choice(target_new_data.shape[0], num_samples_for_testing, replace=False)

    joint_target_samples = np.vstack(
        [shape_into_2d(target_original_data[original_indices]), shape_into_2d(target_new_data[new_indices])]
    )

    if parents_original_data is None:
        return independence_test(joint_target_samples, data_set_indices)
    else:
        parents_new_data: np.ndarray
        joint_parent_data = np.vstack(
            [shape_into_2d(parents_original_data[original_indices]), shape_into_2d(parents_new_data[new_indices])]
        )

        return conditional_independence_test(joint_target_samples, data_set_indices, joint_parent_data)


def distribution_change(
    causal_model: ProbabilisticCausalModel,
    old_data: pd.DataFrame,
    new_data: pd.DataFrame,
    target_node: Any,
    num_samples: int = 2000,
    difference_estimation_func: Callable[[np.ndarray, np.ndarray], float] = auto_estimate_kl_divergence,
    independence_test: Callable[[np.ndarray, np.ndarray], float] = kernel_based,
    conditional_independence_test: Callable[[np.ndarray, np.ndarray, np.ndarray], float] = kernel_based,
    mechanism_change_test_significance_level: float = 0.05,
    mechanism_change_test_fdr_control_method: Optional[str] = "fdr_bh",
    auto_assignment_quality: Optional[AssignmentQuality] = None,
    return_additional_info: bool = False,
    shapley_config: Optional[ShapleyConfig] = None,
    graph_factory: Callable[[Any], DirectedGraph] = nx.DiGraph,
) -> Union[
    Dict[Any, float], Tuple[Dict[Any, float], Dict[Any, bool], ProbabilisticCausalModel, ProbabilisticCausalModel]
]:
    """Attributes the change in the marginal distribution of the target_node to nodes upstream in the causal DAG.

    Note that this method creates two copies of the causal DAG. The causal models of one causal DAG are learned from
    old data and those of another DAG are learned from new data.

    **Research Paper**:
    Kailash Budhathoki, Dominik Janzing, Patrick Bloebaum, Hoiyi Ng. *Why did the distribution change?*. Proceedings
    of The 24th International Conference on Artificial Intelligence and Statistics, PMLR 130:1666-1674, 2021.

    :param causal_model: Reference causal model.
    :param old_data: Joint samples from the 'old' distribution.
    :param new_data: Joint samples from the 'new' distribution.
    :param target_node: Target node of interest for attributing the marginal distribution change.
    :param num_samples: Number of samples used for estimating Shapley values. This can have a significant influence
                        on runtime and accuracy.
    :param difference_estimation_func: Function for quantifying the distribution change. This function should expect
                                       two inputs which represent samples from two different distributions,
                                       e.g. difference in average values.
    :param independence_test: Unconditional independence test. This is used to identify mechanism changes in root nodes.
    :param conditional_independence_test: Conditional independence test. This is used to identify mechanism changes in
                                          non-root nodes.
    :param mechanism_change_test_significance_level: A significance level for rejecting the null hypothesis that the
                                                     causal mechanism of a node has not changed.
    :param mechanism_change_test_fdr_control_method: The false discovery rate control method for mechanism change
                                                     tests. For more options, checkout `statsmodels manual
                                                     <https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html>`_.
    :param auto_assignment_quality: If set to None, the assigned models from the given causal models are used for the
                                    old and new graph. However, they are re-fitted on the given data.
                                    If set to a valid assignment quality, new models are automatically assigned to the
                                    old and new graph based on the respective data.
    :param return_additional_info: If set to True, three additional items are returned: a dictionary indicating
                                   whether each node's mechanism changed, the causal DAG whose causal models are
                                   learned from old data, and the causal DAG whose causal models are learned from new
                                   data.
    :param shapley_config: Configuration for the Shapley estimator.
    :param graph_factory: Allows customization in case a graph class different than networkx.DiGraph should be used.
                          This function *must* copy nodes and edges. Attributes of nodes will be overridden in the copy,
                          so the algorithm is independent of the attribute copy behavior of this factory.
    :return: By default, if `return_additional_info` is set to False, only the dictionary containing contribution of
             each upstream node is returned. If `return_additional_info` is set to True, three additional items are
             returned: a dictionary indicating whether each node's mechanism changed, the causal DAG whose causal models
             learned from old data, and the causal DAG whose causal models are learned from new data.
    """
    causal_graph_old = graph_factory(node_connected_subgraph_view(causal_model.graph, target_node))
    causal_model_old = ProbabilisticCausalModel(causal_graph_old)

    if auto_assignment_quality is None:
        clone_causal_models(causal_model.graph, causal_model_old.graph)
    else:
        assign_causal_mechanisms(causal_model_old, old_data, override_models=True, quality=auto_assignment_quality)

    causal_graph_new = graph_factory(causal_graph_old)
    causal_model_new = ProbabilisticCausalModel(causal_graph_new)
    if auto_assignment_quality is None:
        clone_causal_models(causal_graph_old, causal_model_new.graph)
    else:
        assign_causal_mechanisms(causal_model_new, new_data, override_models=True, quality=auto_assignment_quality)

    mechanism_changes = _fit_accounting_for_mechanism_change(
        causal_model_old,
        causal_model_new,
        old_data[list(causal_graph_old.nodes)],
        new_data[list(causal_graph_new.nodes)],
        independence_test,
        conditional_independence_test,
        mechanism_change_test_significance_level,
        mechanism_change_test_fdr_control_method,
    )

    attributions = distribution_change_of_graphs(
        causal_model_old,
        causal_model_new,
        target_node,
        num_samples,
        difference_estimation_func,
        shapley_config,
        graph_factory,
    )
    if return_additional_info:
        return attributions, mechanism_changes, causal_model_old, causal_model_new
    else:
        return attributions


def distribution_change_of_graphs(
    causal_model_old: ProbabilisticCausalModel,
    causal_model_new: ProbabilisticCausalModel,
    target_node: Any,
    num_samples: int = 2000,
    difference_estimation_func: Callable[[np.ndarray, np.ndarray], float] = auto_estimate_kl_divergence,
    shapley_config: Optional[ShapleyConfig] = None,
    graph_factory: Callable[[Any], DirectedGraph] = nx.DiGraph,
) -> Dict[Any, float]:
    """Attributes the change of the marginal distribution of target_node to upstream nodes based on the distributions
    generated by the 'old' and 'new' causal graphs. These graphs are assumed to represent the same causal structure and
    to be fitted on the respective data.

    Note: This method creates a copy of the given causal models, i.e. the original objects will not be modified.

    Related paper:
    Budhathoki, K., Janzing, D., Bloebaum, P., & Ng, H. (2021). Why did the distribution change?
    arXiv preprint arXiv:2102.13384.

    :param causal_model_old: The ProbabilisticCausalModel fitted on the 'old' data.
    :param causal_model_new: The ProbabilisticCausalModel fitted on the 'new' data.
    :param target_node: Node of interest for attributing the marginal distribution change.
    :param num_samples: Number of samples used for the estimation. This can have a significant influence on the runtime
                        and accuracy.
    :param difference_estimation_func: Function for quantifying the distribution change. This function should expect
                                       two inputs which represent samples from two different distributions. An example
                                       could be the KL divergence.
    :param shapley_config: Config for the Shapley estimator.
    :param graph_factory: Allows customization in case a graph class different than networkx.DiGraph should be used.
                          This function *must* copy nodes and edges. Attributes of nodes will be overridden in the copy,
                          so the algorithm is independent of the attribute copy behavior of this factory.
    :return: A dictionary containing the contributions of upstream nodes to the marginal distribution change in the
             target node.
    """
    validate_causal_dag(causal_model_old.graph)
    validate_causal_dag(causal_model_new.graph)

    return _estimate_marginal_distribution_change(
        ProbabilisticCausalModel(node_connected_subgraph_view(causal_model_old.graph, target_node)),
        ProbabilisticCausalModel(node_connected_subgraph_view(causal_model_new.graph, target_node)),
        target_node,
        num_samples,
        difference_estimation_func,
        shapley_config,
        graph_factory,
    )


def _fit_accounting_for_mechanism_change(
    causal_model_old: ProbabilisticCausalModel,
    causal_model_new: ProbabilisticCausalModel,
    old_data: pd.DataFrame,
    new_data: pd.DataFrame,
    independence_test: Callable[[np.ndarray, np.ndarray], float],
    conditional_independence_test: Callable[[np.ndarray, np.ndarray, np.ndarray], float],
    significance_level: float,
    fdr_control_method: Optional[str],
) -> Dict[Any, bool]:
    mechanism_changed_for_node = _check_significant_mechanism_change(
        causal_model_old.graph,
        old_data,
        new_data,
        independence_test,
        conditional_independence_test,
        significance_level,
        fdr_control_method,
    )

    joint_data = pd.concat([old_data, new_data], ignore_index=True, sort=True)

    for node in causal_model_new.graph.nodes:
        if mechanism_changed_for_node[node]:
            fit_causal_model_of_target(causal_model_old, node, old_data)
            fit_causal_model_of_target(causal_model_new, node, new_data)
        else:
            fit_causal_model_of_target(causal_model_old, node, joint_data)
            fit_causal_model_of_target(causal_model_new, node, joint_data)

    return mechanism_changed_for_node


def _estimate_marginal_distribution_change(
    causal_model_old: ProbabilisticCausalModel,
    causal_model_new: ProbabilisticCausalModel,
    target_node: Any,
    num_samples: int,
    difference_estimation_func: Callable[[np.ndarray, np.ndarray], float],
    shapley_config: Optional[ShapleyConfig],
    graph_factory: Callable[[Any], DirectedGraph],
) -> Dict[Any, float]:
    old_causal_models = [causal_model_old.causal_mechanism(x) for x in sorted(causal_model_old.graph.nodes)]
    new_causal_models = [causal_model_new.causal_mechanism(x) for x in sorted(causal_model_new.graph.nodes)]

    target_samples_old = draw_samples(causal_model_old, num_samples)[target_node].to_numpy()

    def attribution_set_function(subset):
        if np.all(subset == 0):
            return 0

        causal_model = ProbabilisticCausalModel(graph_factory(causal_model_old.graph))
        nodes = sorted(list(causal_model.graph.nodes))

        for i in range(len(old_causal_models)):
            if subset[i] == 1:
                causal_model.set_causal_mechanism(nodes[i], new_causal_models[i])
            else:
                causal_model.set_causal_mechanism(nodes[i], old_causal_models[i])

        for node in causal_model.graph.nodes:
            causal_model.graph.nodes[node][PARENTS_DURING_FIT] = get_ordered_predecessors(causal_model.graph, node)

        target_samples_new = draw_samples(causal_model, num_samples)[target_node].to_numpy()

        return difference_estimation_func(target_samples_old, target_samples_new)

    attributions = estimate_shapley_values(attribution_set_function, len(old_causal_models), shapley_config)

    return {x: attributions[i] for i, x in enumerate(sorted(causal_model_old.graph.nodes))}


def estimate_distribution_change_scores(
    causal_model: ProbabilisticCausalModel,
    original_data: pd.DataFrame,
    new_data: pd.DataFrame,
    difference_estimation_func: Callable[
        [np.ndarray, np.ndarray], Union[np.ndarray, float]
    ] = auto_estimate_kl_divergence,
    max_num_evaluation_samples: int = 1000,
    num_joint_samples: int = 500,
    early_stopping_percentage: float = 0.01,
    independence_test: Callable[[np.ndarray, np.ndarray], float] = kernel_based,
    conditional_independence_test: Callable[[np.ndarray, np.ndarray, np.ndarray], float] = kernel_based,
    mechanism_change_test_significance_level: float = 0.05,
    mechanism_change_test_fdr_control_method: Optional[str] = "fdr_bh",
) -> Dict[Any, float]:
    """Given newly observed and original samples from the joint distribution of the given causal graphical model, this
    method estimates a score for each node that quantifies how much the distribution of the node has changed. For this,
    it first checks whether the underlying causal mechanism has changed at all and, if this is the case, it estimates
    the difference between the new and original distributions. The score is based on the quantity measured by the
    provided difference_estimation_func or 0 if no mechanism change has been detected.

    Note that for each parent sample, num_joint_samples conditional samples are generated based on the original and new
    causal mechanism and evaluated by the given difference_estimation_func function. These results are then averaged
    over multiple different parent samples.

    :param causal_model: The underlying causal model based on the original data.
    :param original_data: Samples from the original data.
    :param new_data: Samples from the new data.
    :param difference_estimation_func: Function for quantifying the distribution change. This function should expect
                                       two inputs which represent samples from two different distributions. An example
                                       could be the KL divergence.
    :param max_num_evaluation_samples: Maximum number of (parent) samples for evaluating the difference in
                                       distributions.
    :param num_joint_samples: Number of samples generated in a node per parent sample.
    :param early_stopping_percentage: If the change in percentage between multiple consecutive runs is below this
                                      threshold, the evaluation stops before evaluating all max_num_evaluation_samples.
    :param independence_test: Unconditional independence test. This is used to identify mechanism changes in root nodes.
    :param conditional_independence_test: Conditional independence test. This is used to identify mechanism changes in
                                          non-root nodes.
    :param mechanism_change_test_significance_level: A significance level for rejecting the null hypothesis that the
                                                     causal mechanism of a node has not changed.
    :param mechanism_change_test_fdr_control_method: The false discovery rate control method for mechanism change
                                                     tests. For more options, checkout `statsmodels manual
                                                     <https://www.statsmodels.org/dev/generated/statsmodels.stats.multitest.multipletests.html>`_.
    :return: A dictionary assining a score to each node in the causal graph.
    """
    validate_causal_dag(causal_model.graph)

    mechanism_changed_for_node = _check_significant_mechanism_change(
        causal_model.graph,
        original_data,
        new_data,
        independence_test,
        conditional_independence_test,
        mechanism_change_test_significance_level,
        mechanism_change_test_fdr_control_method,
    )

    results = {}
    for node in tqdm(
        causal_model.graph.nodes, desc="Estimating mechanism change anomaly scores", position=0, leave=True
    ):
        if mechanism_changed_for_node[node]:
            if is_root_node(causal_model.graph, node):
                results[node] = difference_estimation_func(original_data[node].to_numpy(), new_data[node].to_numpy())
            else:
                parent_nodes = get_ordered_predecessors(causal_model.graph, node)
                results[node] = _estimate_distribution_change_score(
                    original_data[parent_nodes].to_numpy(),
                    new_data[parent_nodes].to_numpy(),
                    new_data[node].to_numpy(),
                    causal_model.causal_mechanism(node),
                    difference_estimation_func,
                    max_num_evaluation_samples,
                    num_joint_samples,
                    early_stopping_percentage,
                )
        else:
            results[node] = 0

    return results


def _check_significant_mechanism_change(
    graph: DirectedGraph,
    old_data: pd.DataFrame,
    new_data: pd.DataFrame,
    independence_test: Callable[[np.ndarray, np.ndarray], float],
    conditional_independence_test: Callable[[np.ndarray, np.ndarray, np.ndarray], float],
    significance_level: float,
    fdr_control_method: Optional[str],
) -> Dict[Any, bool]:
    all_p_values = []
    for node in graph.nodes:
        if is_root_node(graph, node):
            parents_org_data = None
            parents_new_data = None
        else:
            parents_org_data = old_data[get_ordered_predecessors(graph, node)].to_numpy()
            parents_new_data = new_data[get_ordered_predecessors(graph, node)].to_numpy()

        all_p_values.append(
            mechanism_change_test(
                old_data[node].to_numpy(),
                new_data[node].to_numpy(),
                parents_org_data,
                parents_new_data,
                independence_test=independence_test,
                conditional_independence_test=conditional_independence_test,
            )
        )

    if fdr_control_method is None:
        successes = np.array(all_p_values) <= significance_level
    else:
        successes = multipletests(all_p_values, significance_level, method=fdr_control_method)[0]

    return dict(zip(graph.nodes, successes))


def _estimate_distribution_change_score(
    parent_original_data: np.ndarray,
    parent_new_data: np.ndarray,
    target_new_data: np.ndarray,
    causal_model_original: ConditionalStochasticModel,
    difference_estimation_func: Callable[[np.ndarray, np.ndarray], Union[np.ndarray, float]],
    max_num_evaluation_samples: int,
    num_joint_samples: int,
    early_stopping_percentage: float,
) -> float:
    parent_original_data, parent_new_data, target_new_data = shape_into_2d(
        parent_original_data, parent_new_data, target_new_data
    )
    causal_model_new = causal_model_original.clone()
    causal_model_new.fit(X=parent_new_data, Y=target_new_data)

    joint_parent_samples = np.vstack([parent_original_data, parent_new_data])
    joint_parent_samples = joint_parent_samples[
        np.random.choice(
            joint_parent_samples.shape[0], min(joint_parent_samples.shape[0], max_num_evaluation_samples), replace=False
        )
    ]

    result = 0
    run = 0
    for joint_parent_sample in joint_parent_samples:
        old_result = result

        samples = repmat(joint_parent_sample, num_joint_samples, 1)
        result += difference_estimation_func(
            causal_model_original.draw_samples(samples), causal_model_new.draw_samples(samples)
        )

        run += 1
        if old_result != 0 and (1 - result / old_result) <= early_stopping_percentage:
            # If the relative change of the score is less than the given threshold, we stop the estimation early.
            _logger.info(
                "Early stopping: Result only changed by %f percent and a threshold of %f is set."
                % (1 - result / old_result, early_stopping_percentage)
            )
            break

    return result / run



import itertools
import logging
import re

import networkx as nx

from dowhy.utils.api import parse_state
from dowhy.utils.graph_operations import daggity_to_dot


class CausalGraph:

    """Class for creating and modifying the causal graph.

    Accepts a graph string (or a text file) in gml format (preferred) and dot format. Graphviz-like attributes can be set for edges and nodes. E.g. style="dashed" as an edge attribute ensures that the edge is drawn with a dashed line.

     If a graph string is not given, names of treatment, outcome, and confounders, instruments and effect modifiers (if any) can be provided to create the graph.
    """

    def __init__(
        self,
        treatment_name,
        outcome_name,
        graph=None,
        common_cause_names=None,
        instrument_names=None,
        effect_modifier_names=None,
        mediator_names=None,
        observed_node_names=None,
        missing_nodes_as_confounders=False,
    ):
        self.treatment_name = parse_state(treatment_name)
        self.outcome_name = parse_state(outcome_name)
        instrument_names = parse_state(instrument_names)
        common_cause_names = parse_state(common_cause_names)
        effect_modifier_names = parse_state(effect_modifier_names)
        mediator_names = parse_state(mediator_names)
        self.logger = logging.getLogger(__name__)

        # re.sub only takes string parameter so the first if is to avoid error
        # if the input is a text file, convert the contained data into string
        if isinstance(graph, str) and re.match(r".*\.txt", str(graph)):
            text_file = open(graph, "r")
            graph = text_file.read()
            text_file.close()

        if isinstance(graph, str) and re.match(r"^dag", graph):  # Convert daggity output to dot format
            graph = daggity_to_dot(graph)

        if isinstance(graph, str):
            graph = graph.replace("\n", " ")

        if graph is None:
            self._graph = nx.DiGraph()
            self._graph = self.build_graph(common_cause_names, instrument_names, effect_modifier_names, mediator_names)
        elif re.match(r".*\.dot", graph):
            # load dot file
            try:
                import pygraphviz as pgv

                self._graph = nx.DiGraph(nx.drawing.nx_agraph.read_dot(graph))
            except Exception as e:
                self.logger.error("Pygraphviz cannot be loaded. " + str(e) + "\nTrying pydot...")
                try:
                    import pydot

                    self._graph = nx.DiGraph(nx.drawing.nx_pydot.read_dot(graph))
                except Exception as e:
                    self.logger.error("Error: Pydot cannot be loaded. " + str(e))
                    raise e
        elif re.match(r".*\.gml", graph):
            self._graph = nx.DiGraph(nx.read_gml(graph))
        elif re.match(r".*graph\s*\{.*\}\s*", graph):
            try:
                import pygraphviz as pgv

                self._graph = pgv.AGraph(graph, strict=True, directed=True)
                self._graph = nx.drawing.nx_agraph.from_agraph(self._graph)
            except Exception as e:
                self.logger.error("Error: Pygraphviz cannot be loaded. " + str(e) + "\nTrying pydot ...")
                try:
                    import pydot

                    P_list = pydot.graph_from_dot_data(graph)
                    self._graph = nx.drawing.nx_pydot.from_pydot(P_list[0])
                except Exception as e:
                    self.logger.error("Error: Pydot cannot be loaded. " + str(e))
                    raise e
        elif re.match(".*graph\s*\[.*\]\s*", graph):
            self._graph = nx.DiGraph(nx.parse_gml(graph))
        else:
            self.logger.error("Error: Please provide graph (as string or text file) in dot or gml format.")
            self.logger.error("Error: Incorrect graph format")
            raise ValueError
        if missing_nodes_as_confounders:
            self._graph = self.add_missing_nodes_as_common_causes(observed_node_names)
        # Adding node attributes
        self._graph = self.add_node_attributes(observed_node_names)

    def view_graph(self, layout="dot", size=(8, 6), file_name="causal_model"):
        out_filename = "{}.png".format(file_name)
        try:
            import pygraphviz as pgv

            agraph = nx.drawing.nx_agraph.to_agraph(self._graph)
            agraph.graph_attr.update(size="{},{}!".format(size[0], size[0]))
            agraph.draw(out_filename, format="png", prog=layout)
        except:
            self.logger.warning(
                "Warning: Pygraphviz cannot be loaded. Check that graphviz and pygraphviz are installed."
            )
            self.logger.info("Using Matplotlib for plotting")
            import matplotlib.pyplot as plt

            plt.figure(figsize=size)
            solid_edges = [(n1, n2) for n1, n2, e in self._graph.edges(data=True) if "style" not in e]
            dashed_edges = [
                (n1, n2) for n1, n2, e in self._graph.edges(data=True) if ("style" in e and e["style"] == "dashed")
            ]
            plt.clf()

            pos = nx.layout.shell_layout(self._graph)
            nx.draw_networkx_nodes(self._graph, pos, node_color="yellow", node_size=400)
            nx.draw_networkx_edges(self._graph, pos, edgelist=solid_edges, arrowstyle="-|>", arrowsize=12)
            nx.draw_networkx_edges(
                self._graph, pos, edgelist=dashed_edges, arrowstyle="-|>", style="dashed", arrowsize=12
            )

            labels = nx.draw_networkx_labels(self._graph, pos)

            plt.axis("off")
            plt.savefig(out_filename)
            plt.draw()

    def build_graph(self, common_cause_names, instrument_names, effect_modifier_names, mediator_names):
        """Creates nodes and edges based on variable names and their semantics.

        Currently only considers the graphical representation of "direct" effect modifiers. Thus, all effect modifiers are assumed to be "direct" unless otherwise expressed using a graph. Based on the taxonomy of effect modifiers by VanderWheele and Robins: "Four types of effect modification: A classification based on directed acyclic graphs. Epidemiology. 2007."
        """

        for treatment in self.treatment_name:
            self._graph.add_node(treatment, observed="yes", penwidth=2)
        for outcome in self.outcome_name:
            self._graph.add_node(outcome, observed="yes", penwidth=2)
        for treatment, outcome in itertools.product(self.treatment_name, self.outcome_name):
            # adding penwidth to make the edge bold
            self._graph.add_edge(treatment, outcome, penwidth=2)

        # Adding common causes
        if common_cause_names is not None:
            for node_name in common_cause_names:
                for treatment, outcome in itertools.product(self.treatment_name, self.outcome_name):
                    self._graph.add_node(node_name, observed="yes")
                    self._graph.add_edge(node_name, treatment)
                    self._graph.add_edge(node_name, outcome)

        # Adding instruments
        if instrument_names:
            if type(instrument_names[0]) != tuple:
                if len(self.treatment_name) > 1:
                    self.logger.info("Assuming Instrument points to all treatments! Use tuples for more granularity.")
                for instrument, treatment in itertools.product(instrument_names, self.treatment_name):
                    self._graph.add_node(instrument, observed="yes")
                    self._graph.add_edge(instrument, treatment)
            else:
                for instrument, treatment in itertools.product(instrument_names):
                    self._graph.add_node(instrument, observed="yes")
                    self._graph.add_edge(instrument, treatment)

        # Adding effect modifiers
        if effect_modifier_names is not None:
            for node_name in effect_modifier_names:
                if node_name not in common_cause_names:
                    for outcome in self.outcome_name:
                        self._graph.add_node(node_name, observed="yes")
                        # Assuming the simple form of effect modifier
                        # that directly causes the outcome.
                        self._graph.add_edge(node_name, outcome)
                        # self._graph.add_edge(node_name, outcome, style = "dotted", headport="s", tailport="n")
                        # self._graph.add_edge(outcome, node_name, style = "dotted", headport="n", tailport="s") # TODO make the ports more general so that they apply not just to top-bottom node configurations
        if mediator_names is not None:
            for node_name in mediator_names:
                for treatment, outcome in itertools.product(self.treatment_name, self.outcome_name):
                    self._graph.add_node(node_name, observed="yes")
                    self._graph.add_edge(treatment, node_name)
                    self._graph.add_edge(node_name, outcome)
        return self._graph

    def add_node_attributes(self, observed_node_names):
        for node_name in self._graph:
            if node_name in observed_node_names:
                self._graph.nodes[node_name]["observed"] = "yes"
            else:
                self._graph.nodes[node_name]["observed"] = "no"
        return self._graph

    def add_missing_nodes_as_common_causes(self, observed_node_names):
        # Adding columns in the dataframe as confounders that were not in the graph
        for node_name in observed_node_names:
            if node_name not in self._graph:
                self._graph.add_node(node_name, observed="yes")
                for treatment_outcome_node in self.treatment_name + self.outcome_name:
                    self._graph.add_edge(node_name, treatment_outcome_node)
        return self._graph

    def add_unobserved_common_cause(self, observed_node_names, color="gray"):
        # Adding unobserved confounders
        current_common_causes = self.get_common_causes(self.treatment_name, self.outcome_name)
        create_new_common_cause = True
        for node_name in current_common_causes:
            if self._graph.nodes[node_name]["observed"] == "no":
                create_new_common_cause = False
        if create_new_common_cause:
            uc_label = "Unobserved Confounders"
            self._graph.add_node("U", label=uc_label, observed="no", color=color, style="filled", fillcolor=color)
            for node in self.treatment_name + self.outcome_name:
                self._graph.add_edge("U", node)
            self.logger.info(
                'If this is observed data (not from a randomized experiment), there might always be missing confounders. Adding a node named "Unobserved Confounders" to reflect this.'
            )
        return self._graph

    def get_unconfounded_observed_subgraph(self):
        observed_nodes = [node for node in self._graph.nodes() if self._graph.nodes[node]["observed"] == "yes"]
        return self._graph.subgraph(observed_nodes)

    def do_surgery(
        self,
        node_names,
        remove_outgoing_edges=False,
        remove_incoming_edges=False,
        target_node_names=None,
        remove_only_direct_edges_to_target=False,
    ):
        """Method to create a new graph based on the concept of do-surgery.

        :param node_names: focal nodes for the surgery
        :param remove_outgoing_edges: whether to remove outgoing edges from the focal nodes
        :param remove_incoming_edges: whether to remove incoming edges to the focal nodes
        :param target_node_names: target nodes (optional) for the surgery, only used when remove_only_direct_edges_to_target is True
        :param remove_only_direct_edges_to_target: whether to remove only the direct edges from focal nodes to the target nodes

        :returns: a new networkx graph after the specified removal of edges
        """

        node_names = parse_state(node_names)
        new_graph = self._graph.copy()
        for node_name in node_names:
            if remove_outgoing_edges:
                if remove_only_direct_edges_to_target:
                    new_graph.remove_edges_from([(node_name, v) for v in target_node_names])
                else:
                    children = new_graph.successors(node_name)
                    edges_bunch = [(node_name, child) for child in children]
                    new_graph.remove_edges_from(edges_bunch)
            if remove_incoming_edges:
                # removal of only direct edges wrt a target is not implemented for incoming edges
                parents = new_graph.predecessors(node_name)
                edges_bunch = [(parent, node_name) for parent in parents]
                new_graph.remove_edges_from(edges_bunch)
        return new_graph

    def get_causes(self, nodes, remove_edges=None):
        nodes = parse_state(nodes)
        new_graph = None
        if remove_edges is not None:
            new_graph = self._graph.copy()  # caution: shallow copy of the attributes
            sources = parse_state(remove_edges["sources"])
            targets = parse_state(remove_edges["targets"])
            for s in sources:
                for t in targets:
                    new_graph.remove_edge(s, t)
        causes = set()
        for v in nodes:
            causes = causes.union(self.get_ancestors(v, new_graph=new_graph))
        return causes

    def check_dseparation(self, nodes1, nodes2, nodes3, new_graph=None, dseparation_algo="default"):
        if dseparation_algo == "default":
            if new_graph is None:
                new_graph = self._graph
            dseparated = nx.algorithms.d_separated(new_graph, set(nodes1), set(nodes2), set(nodes3))
        else:
            raise ValueError(f"{dseparation_algo} method for d-separation not supported.")
        return dseparated

    def check_valid_backdoor_set(
        self, nodes1, nodes2, nodes3, backdoor_paths=None, new_graph=None, dseparation_algo="default"
    ):
        """Assume that the first parameter (nodes1) is the treatment,
        the second is the outcome, and the third is the candidate backdoor set
        """
        # also return the number of backdoor paths blocked by observed nodes
        if dseparation_algo == "default":
            if new_graph is None:
                # Assume that nodes1 is the treatment
                new_graph = self.do_surgery(nodes1, remove_outgoing_edges=True)
            dseparated = nx.algorithms.d_separated(new_graph, set(nodes1), set(nodes2), set(nodes3))
        elif dseparation_algo == "naive":
            # ignores new_graph parameter, always uses self._graph
            if backdoor_paths is None:
                backdoor_paths = self.get_backdoor_paths(nodes1, nodes2)
            dseparated = all([self.is_blocked(path, nodes3) for path in backdoor_paths])
        else:
            raise ValueError(f"{dseparation_algo} method for d-separation not supported.")
        return {"is_dseparated": dseparated}

    def get_backdoor_paths(self, nodes1, nodes2):
        paths = []
        undirected_graph = self._graph.to_undirected()
        nodes12 = set(nodes1).union(nodes2)
        for node1 in nodes1:
            for node2 in nodes2:
                backdoor_paths = [
                    pth
                    for pth in nx.all_simple_paths(undirected_graph, source=node1, target=node2)
                    if self._graph.has_edge(pth[1], pth[0])
                ]
                # remove paths that have nodes1\node1 or nodes2\node2 as intermediate nodes
                filtered_backdoor_paths = [pth for pth in backdoor_paths if len(nodes12.intersection(pth[1:-1])) == 0]
                paths.extend(filtered_backdoor_paths)
        self.logger.debug("Backdoor paths: " + str(paths))
        return paths

    def is_blocked(self, path, conditioned_nodes):
        """Uses d-separation criteria to decide if conditioned_nodes block given path."""

        blocked_by_conditioning = False
        has_unconditioned_collider = False

        for i in range(len(path) - 2):
            if self._graph.has_edge(path[i], path[i + 1]) and self._graph.has_edge(
                path[i + 2], path[i + 1]
            ):  # collider
                collider_descendants = nx.descendants(self._graph, path[i + 1])
                if path[i + 1] not in conditioned_nodes and all(
                    cdesc not in conditioned_nodes for cdesc in collider_descendants
                ):
                    has_unconditioned_collider = True
            else:  # chain or fork
                if path[i + 1] in conditioned_nodes:
                    blocked_by_conditioning = True
                    break
        if blocked_by_conditioning:
            return True
        elif has_unconditioned_collider:
            return True
        else:
            return False

    def get_common_causes(self, nodes1, nodes2):
        """
        Assume that nodes1 causes nodes2 (e.g., nodes1 are the treatments and nodes2 are the outcomes)
        """
        # TODO Refactor to remove this from here and only implement this logic in causalIdentifier. Unnecessary assumption of nodes1 to be causing nodes2.
        nodes1 = parse_state(nodes1)
        nodes2 = parse_state(nodes2)
        causes_1 = set()
        causes_2 = set()
        for node in nodes1:
            causes_1 = causes_1.union(self.get_ancestors(node))
        for node in nodes2:
            # Cannot simply compute ancestors, since that will also include nodes1 and its parents (e.g. instruments)
            parents_2 = self.get_parents(node)
            for parent in parents_2:
                if parent not in nodes1:
                    causes_2 = causes_2.union(
                        set(
                            [
                                parent,
                            ]
                        )
                    )
                    causes_2 = causes_2.union(self.get_ancestors(parent))
        return list(causes_1.intersection(causes_2))

    def get_effect_modifiers(self, nodes1, nodes2):
        modifiers = set()
        for node in nodes2:
            modifiers = modifiers.union(self.get_ancestors(node))
        modifiers = modifiers.difference(nodes1)
        for node in nodes1:
            modifiers = modifiers.difference(self.get_ancestors(node))
        # removing all mediators
        for node1 in nodes1:
            for node2 in nodes2:
                all_directed_paths = nx.all_simple_paths(self._graph, node1, node2)
                for path in all_directed_paths:
                    modifiers = modifiers.difference(path)
        return list(modifiers)

    def get_parents(self, node_name):
        return set(self._graph.predecessors(node_name))

    def get_ancestors(self, node_name, new_graph=None):
        if new_graph is None:
            graph = self._graph
        else:
            graph = new_graph
        return set(nx.ancestors(graph, node_name))

    def get_descendants(self, nodes):
        descendants = set()
        for node_name in nodes:
            descendants = descendants.union(set(nx.descendants(self._graph, node_name)))
        return descendants

    def all_observed(self, node_names):
        for node_name in node_names:
            if self._graph.nodes[node_name]["observed"] != "yes":
                return False

        return True

    def get_all_nodes(self, include_unobserved=True):
        nodes = self._graph.nodes
        if not include_unobserved:
            nodes = set(self.filter_unobserved_variables(nodes))

        return nodes

    def filter_unobserved_variables(self, node_names):
        observed_node_names = list()
        for node_name in node_names:
            if self._graph.nodes[node_name]["observed"] == "yes":
                observed_node_names.append(node_name)

        return observed_node_names

    def get_instruments(self, treatment_nodes, outcome_nodes):
        treatment_nodes = parse_state(treatment_nodes)
        outcome_nodes = parse_state(outcome_nodes)
        parents_treatment = set()
        for node in treatment_nodes:
            parents_treatment = parents_treatment.union(self.get_parents(node))
        g_no_parents_treatment = self.do_surgery(treatment_nodes, remove_incoming_edges=True)
        ancestors_outcome = set()
        for node in outcome_nodes:
            ancestors_outcome = ancestors_outcome.union(nx.ancestors(g_no_parents_treatment, node))
        # [TODO: double check these work with multivariate implementation:]
        # Exclusion
        candidate_instruments = parents_treatment.difference(ancestors_outcome)
        self.logger.debug("Candidate instruments after satisfying exclusion: %s", candidate_instruments)
        # As-if-random setup
        children_causes_outcome = [nx.descendants(g_no_parents_treatment, v) for v in ancestors_outcome]
        children_causes_outcome = set([item for sublist in children_causes_outcome for item in sublist])

        # As-if-random
        instruments = candidate_instruments.difference(children_causes_outcome)
        self.logger.debug("Candidate instruments after satisfying exclusion and as-if-random: %s", instruments)
        return list(instruments)

    def get_all_directed_paths(self, nodes1, nodes2):
        """Get all directed paths between sets of nodes.

        Currently only supports singleton sets.
        """
        node1 = nodes1[0]
        node2 = nodes2[0]
        # convert the outputted generator into a list
        return [p for p in nx.all_simple_paths(self._graph, source=node1, target=node2)]

    def has_directed_path(self, nodes1, nodes2):
        """Checks if there is any directed path between two sets of nodes.

        Currently only supports singleton sets.
        """
        # dpaths = self.get_all_directed_paths(nodes1, nodes2)
        # return len(dpaths) > 0
        return nx.has_path(self._graph, nodes1[0], nodes2[0])

    def get_adjacency_matrix(self, *args, **kwargs):
        """
        Get adjacency matrix from the networkx graph

        """
        return nx.convert_matrix.to_numpy_matrix(self._graph, *args, **kwargs)

    def check_valid_frontdoor_set(
        self, nodes1, nodes2, candidate_nodes, frontdoor_paths=None, new_graph=None, dseparation_algo="default"
    ):
        """Check if valid the frontdoor variables for set of treatments, nodes1 to set of outcomes, nodes2."""
        # Condition 1: node 1 ---> node 2 is intercepted by candidate_nodes
        if dseparation_algo == "default":
            if new_graph is None:
                new_graph = self._graph
            dseparated = nx.algorithms.d_separated(new_graph, set(nodes1), set(nodes2), set(candidate_nodes))
        elif dseparation_algo == "naive":
            if frontdoor_paths is None:
                frontdoor_paths = self.get_all_directed_paths(nodes1, nodes2)

            dseparated = all([self.is_blocked(path, candidate_nodes) for path in frontdoor_paths])
        else:
            raise ValueError(f"{dseparation_algo} method for d-separation not supported.")
        return dseparated

    def check_valid_mediation_set(self, nodes1, nodes2, candidate_nodes, mediation_paths=None):
        """Check if candidate nodes are valid mediators for set of treatments, nodes1 to set of outcomes, nodes2."""
        if mediation_paths is None:
            mediation_paths = self.get_all_directed_paths(nodes1, nodes2)

        is_mediator = any([self.is_blocked(path, candidate_nodes) for path in mediation_paths])
        return is_mediator



import numpy as np
import pandas as pd


class DataGeneratingProcess:

    DEFAULT_PERCENTILE = 0.9

    def __init__(self, **kwargs):
        """
        Base class for implementation of data generating process.

        Subclasses implement functions that create various data generating processes. All data generating processes are in the package "dowhy.utils.dgps".
        """
        self.treatment = kwargs["treatment"]
        self.outcome = kwargs["outcome"]
        self.confounder = kwargs["confounder"]
        self.effect_modifier = kwargs["effect_modifier"]
        self.weights = kwargs.pop("weights", {})
        self.bias = kwargs.pop("bias", {})
        self.seed = kwargs.pop("seed", None)
        self.treatment_is_binary = kwargs.pop("treatment_is_binary", False)
        if self.treatment_is_binary:
            self.percentile = kwargs.pop("percentile", DataGeneratingProcess.DEFAULT_PERCENTILE)
        elif kwargs.pop("percentile", None) is not None:
            raise ValueError("Cannot use percentile, if the input is non-binary")
        else:
            self.percentile = "NA"
        self.true_value = None
        if self.seed is not None:
            np.random.seed(self.seed)

    def generate_data(self):
        raise NotImplementedError()

    def generation_process(self):
        raise NotImplementedError()

    def convert_to_binary(self, data, deterministic=False):
        if deterministic:
            precentile = np.percentile(data, self.percentile, axis=0)
            binary_treat_value = data <= precentile
        else:
            temp = data.argsort(axis=0)
            ranks = np.empty_like(temp)
            ranks[temp[:, 0], 0] = np.arange(data.shape[0])
            prob_t = ranks / data.shape[0]
            # Generating data with equal 0 and 1 (since ranks are uniformly distributed)
            binary_treat_value = np.random.binomial(1, prob_t[:, 0], data.shape[0])

            # Flipping some values
            if self.percentile >= 0.5:
                mask = np.random.binomial(
                    1, (1 - self.percentile) * 2, len(binary_treat_value[binary_treat_value == 1])
                )
                binary_treat_value[binary_treat_value == 1] = mask * binary_treat_value[binary_treat_value == 1]
            else:
                mask = np.random.binomial(1, 1 - self.percentile * 2, len(binary_treat_value[binary_treat_value == 0]))
                binary_treat_value[binary_treat_value == 0] = mask + binary_treat_value[binary_treat_value == 0]
            binary_treat_value = binary_treat_value[:, np.newaxis]

        return binary_treat_value.astype(float)

    def __str__(self):
        rep = """
        treatment:{}
        outcome:{}
        confounder: {}
        effect_modifier: {}
        weights: {}
        bias: {}
        seed: {}
        treatment_is_binary: {}
        percentile: {}
        """.format(
            self.treatment,
            self.outcome,
            self.confounder,
            self.effect_modifier,
            self.weights,
            self.bias,
            self.seed,
            self.treatment_is_binary,
            self.percentile,
        )

        return rep



import logging
import random
from enum import Enum
from typing import List, Union

import numpy as np
import scipy.stats as st

from dowhy.utils.api import parse_state


class SignificanceTestType(Enum):

    AUTO = "auto"
    BOOTSTRAP = "bootstrap"
    NORMAL = "normal_test"


logger = logging.getLogger(__name__)


class CausalRefuter:

    """Base class for different refutation methods.

    Subclasses implement specific refutations methods.

    # todo: add docstring for common parameters here and remove from child refuter classes

    This class is for backwards compatibility with CausalModel
    Will be deprecated in the future in favor of function call refute_method_name() functions

    """

    # Default value for the number of simulations to be conducted
    DEFAULT_NUM_SIMULATIONS = 100
    PROGRESS_BAR_COLOR = "green"

    def __init__(self, data, identified_estimand, estimate, **kwargs):
        self._data = data
        self._target_estimand = identified_estimand
        self._estimate = estimate
        self._treatment_name = self._target_estimand.treatment_variable
        self._outcome_name = self._target_estimand.outcome_variable
        self._random_seed = None

        # joblib params for parallel processing
        self._n_jobs = kwargs.pop("n_jobs", None)
        self._verbose = kwargs.pop("verbose", 0)

        if "random_seed" in kwargs:
            self._random_seed = kwargs["random_seed"]
            np.random.seed(self._random_seed)

        # Concatenate the confounders, instruments and effect modifiers
        try:
            self._variables_of_interest = (
                self._target_estimand.get_backdoor_variables()
                + self._target_estimand.instrumental_variables
                + self._estimate.estimator._effect_modifier_names
            )
        except AttributeError as attr_error:
            logger.error(attr_error)

    def choose_variables(self, required_variables):
        return choose_variables(required_variables, self._variables_of_interest)

    def test_significance(self, estimate, simulations, test_type="auto", significance_level=0.05):
        return test_significance(estimate, simulations, SignificanceTestType(test_type), significance_level)

    def perform_bootstrap_test(self, estimate, simulations):
        return perform_bootstrap_test(estimate, simulations)

    def perform_normal_distribution_test(self, estimate, simulations):
        return perform_normal_distribution_test(estimate, simulations)

    def refute_estimate(self, show_progress_bar=False):
        raise NotImplementedError


def choose_variables(required_variables: Union[bool, int, list], variables_of_interest: List):
    """
    This method provides a way to choose the confounders whose values we wish to
    modify for finding its effect on the ability of the treatment to affect the outcome.
    """

    invert = None

    if required_variables is False:

        logger.info(
            "All variables required: Running bootstrap adding noise to confounders, instrumental variables and effect modifiers."
        )
        return None

    elif required_variables is True:

        logger.info(
            "All variables required: Running bootstrap adding noise to confounders, instrumental variables and effect modifiers."
        )
        return variables_of_interest

    elif type(required_variables) is int:

        if len(variables_of_interest) < required_variables:
            logger.error(
                "Too many variables passed.\n The number of  variables is: {}.\n The number of variables passed: {}".format(
                    len(variables_of_interest), required_variables
                )
            )
            raise ValueError(
                "The number of variables in the required_variables is greater than the number of confounders, instrumental variables and effect modifiers"
            )
        else:
            # Shuffle the confounders
            return random.sample(variables_of_interest, required_variables)

    elif type(required_variables) is list:

        # Check if all are select or deselect variables
        if all(variable[0] == "-" for variable in required_variables):
            invert = True
            required_variables = [variable[1:] for variable in required_variables]
        elif all(variable[0] != "-" for variable in required_variables):
            invert = False
        else:
            logger.error("{} has both select and delect variables".format(required_variables))
            raise ValueError(
                "It appears that there are some select and deselect variables. Note you can either select or delect variables at a time, but not both"
            )

        # Check if all the required_variables belong to confounders, instrumental variables or effect
        if set(required_variables) - set(variables_of_interest) != set([]):
            logger.error(
                "{} are not confounder, instrumental variable or effect modifier".format(
                    list(set(required_variables) - set(variables_of_interest))
                )
            )
            raise ValueError(
                "At least one of required_variables is not a valid variable name, or it is not a confounder, instrumental variable or effect modifier"
            )

        if invert is False:
            return required_variables
        elif invert is True:
            return list(set(variables_of_interest) - set(required_variables))


def perform_bootstrap_test(estimate, simulations: List):
    # Get the number of simulations
    num_simulations = len(simulations)
    # Sort the simulations
    simulations.sort()
    # Obtain the median value
    median_refute_values = simulations[int(num_simulations / 2)]

    # Performing a two sided test
    if estimate.value > median_refute_values:
        # np.searchsorted tells us the index if it were a part of the array
        # We select side to be left as we want to find the first value that matches
        estimate_index = np.searchsorted(simulations, estimate.value, side="left")
        # We subtact 1 as we are finding the value from the right tail
        p_value = 1 - (estimate_index / num_simulations)
    else:
        # We take the side to be right as we want to find the last index that matches
        estimate_index = np.searchsorted(simulations, estimate.value, side="right")
        # We get the probability with respect to the left tail.
        p_value = estimate_index / num_simulations
    # return twice the determined quantile as this is a two sided test
    return 2 * p_value


def perform_normal_distribution_test(estimate, simulations: List):
    # Get the mean for the simulations
    mean_refute_values = np.mean(simulations)
    # Get the standard deviation for the simulations
    std_dev_refute_values = np.std(simulations)
    # Get the Z Score [(val - mean)/ std_dev ]
    z_score = (estimate.value - mean_refute_values) / std_dev_refute_values

    if z_score > 0:  # Right Tail
        p_value = 1 - st.norm.cdf(z_score)
    else:  # Left Tail
        p_value = st.norm.cdf(z_score)

    return p_value


def test_significance(
    estimate,
    simulations: List,
    test_type: SignificanceTestType = SignificanceTestType.AUTO,
    significance_level: float = 0.85,
):
    """Tests the statistical significance of the estimate obtained to the simulations produced by a refuter.

    The basis behind using the sample statistics of the refuter when we are in fact testing the estimate,
    is due to the fact that, we would ideally expect them to follow the same distribition.

    For refutation tests (e.g., placebo refuters), consider the null distribution as a distribution of effect
    estimates over multiple simulations with placebo treatment, and compute how likely the true estimate (e.g.,
    zero for placebo test) is under the null. If the probability of true effect estimate is lower than the
    p-value, then estimator method fails the test.

    For sensitivity analysis tests (e.g., bootstrap, subset or common cause refuters), the null distribution captures
    the distribution of effect estimates under the "true" dataset (e.g., with an additional confounder or different
    sampling), and we compute the probability of the obtained estimate under this distribution. If the probability is
    lower than the p-value, then the estimator method fails the test.

    Null Hypothesis- The estimate is a part of the distribution
    Alternative Hypothesis- The estimate does not fall in the distribution.

    :param 'estimate': CausalEstimate
        The estimate obtained from the estimator for the original data.
    :param 'simulations': np.array
        An array containing the result of the refuter for the simulations
    :param 'test_type': string, default 'auto'
        The type of test the user wishes to perform.
    :param 'significance_level': float, default 0.05
        The significance level for the statistical test

    :returns: significance_dict: Dict
        A Dict containing the p_value and a boolean that indicates if the result is statistically significant
    """
    # Initializing the p_value
    p_value = 0

    if test_type == SignificanceTestType.AUTO:
        num_simulations = len(simulations)
        if num_simulations >= 100:  # Bootstrapping
            logger.info(
                "Making use of Bootstrap as we have more than 100 examples.\n \
            Note: The greater the number of examples, the more accurate are the confidence estimates"
            )

            # Perform Bootstrap Significance Test with the original estimate and the set of refutations
            p_value = perform_bootstrap_test(estimate, simulations)

        else:
            logger.warning(
                "We assume a Normal Distribution as the sample has less than 100 examples.\n \
            Note: The underlying distribution may not be Normal. We assume that it approaches normal with the increase in sample size."
            )

            # Perform Normal Tests of Significance with the original estimate and the set of refutations
            p_value = perform_normal_distribution_test(estimate, simulations)

    elif test_type == SignificanceTestType.BOOTSTRAP:
        logger.info(
            "Performing Bootstrap Test with {} samples\n \
        Note: The greater the number of examples, the more accurate are the confidence estimates".format(
                len(simulations)
            )
        )

        # Perform Bootstrap Significance Test with the original estimate and the set of refutations
        p_value = perform_bootstrap_test(estimate, simulations)

    elif test_type == SignificanceTestType.NORMAL:
        logger.info(
            "Performing Normal Test with {} samples\n \
        Note: We assume that the underlying distribution is Normal.".format(
                len(simulations)
            )
        )

        # Perform Normal Tests of Significance with the original estimate and the set of refutations
        p_value = perform_normal_distribution_test(estimate, simulations)

    significance_dict = {"p_value": p_value, "is_statistically_significant": p_value <= significance_level}

    return significance_dict


class CausalRefutation:
    """Class for storing the result of a refutation method."""

    def __init__(self, estimated_effect, new_effect, refutation_type):
        self.estimated_effect = estimated_effect
        self.new_effect = new_effect
        self.refutation_type = refutation_type

        self.refutation_result = None

    def add_significance_test_results(self, refutation_result):
        self.refutation_result = refutation_result

    def add_refuter(self, refuter_instance):
        self.refuter = refuter_instance

    def interpret(self, method_name=None, **kwargs):
        """Interpret the refutation results.

        :param method_name: Method used (string) or a list of methods. If None, then the default for the specific refuter is used.

        :returns: None

        """
        if method_name is None:
            method_name = self.refuter.interpret_method
        method_name_arr = parse_state(method_name)
        import dowhy.interpreters as interpreters

        for method in method_name_arr:
            interpreter = interpreters.get_class_object(method)
            interpreter(self, **kwargs).interpret()

    def __str__(self):
        if self.refutation_result is None:
            return "{0}\nEstimated effect:{1}\nNew effect:{2}\n".format(
                self.refutation_type, self.estimated_effect, self.new_effect
            )
        else:
            return "{0}\nEstimated effect:{1}\nNew effect:{2}\np value:{3}\n".format(
                self.refutation_type, self.estimated_effect, self.new_effect, self.refutation_result["p_value"]
            )



import numpy as np
import pytest
from flaky import flaky
from numpy.matlib import repmat
from pytest import approx

from dowhy.gcm.ml import (
    create_hist_gradient_boost_classifier,
    create_hist_gradient_boost_regressor,
    create_linear_regressor,
    create_logistic_regression_classifier,
)
from dowhy.gcm.stats import estimate_ftest_pvalue, marginal_expectation, quantile_based_fwer
from dowhy.gcm.util.general import geometric_median


@flaky(max_runs=5)
def test_estimate_geometric_median():
    a = np.random.normal(10, 1, 100)
    a = np.hstack([a, np.random.normal(10000, 1, 20)])
    b = np.random.normal(-5, 1, 100)
    b = np.hstack([b, np.random.normal(-10000, 1, 20)])

    gm = geometric_median(np.vstack([a, b]).T)

    assert gm[0] == approx(10, abs=0.5)
    assert gm[1] == approx(-5, abs=0.5)


def test_quantile_based_fwer():
    p_values = np.array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1])
    assert quantile_based_fwer(p_values, quantile=0.5) == 0.055 / 0.5
    assert quantile_based_fwer(p_values, quantile=0.25) == 0.0325 / 0.25
    assert quantile_based_fwer(p_values, quantile=0.75) == 0.0775 / 0.75

    assert (
        quantile_based_fwer(np.array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1, 1]), quantile=0.5)
        == 0.06 / 0.5
    )
    assert quantile_based_fwer(np.array([0.9, 0.95, 1]), quantile=0.5) == 1
    assert quantile_based_fwer(np.array([0, 0, 0]), quantile=0.5) == 0
    assert quantile_based_fwer(np.array([0.33]), quantile=0.5) == 0.33


def test_given_p_values_with_nans_when_using_quantile_based_fwer_then_ignores_the_nan_values():
    p_values = np.array([0.01, np.nan, 0.02, 0.03, 0.04, 0.05, np.nan, 0.06, 0.07, 0.08, 0.09, 0.1])
    assert quantile_based_fwer(p_values, quantile=0.5) == 0.055 / 0.5


def test_quantile_based_fwer_scaling():
    p_values = np.array([0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1])
    p_values_scaling = np.array([2, 2, 1, 2, 1, 3, 1, 2, 4, 1])

    assert quantile_based_fwer(p_values, p_values_scaling, quantile=0.5) == approx(0.15)
    assert quantile_based_fwer(p_values, p_values_scaling, quantile=0.25) == approx(0.17)
    assert quantile_based_fwer(p_values, p_values_scaling, quantile=0.75) == approx(0.193, abs=0.001)


def test_quantile_based_fwer_raises_error():
    with pytest.raises(ValueError):
        assert quantile_based_fwer(np.array([0.1, 0.5, 1]), quantile=0)

    with pytest.raises(ValueError):
        assert quantile_based_fwer(np.array([0.1, 0.5, 1]), np.array([1, 2]), quantile=0.1)

    with pytest.raises(ValueError):
        assert quantile_based_fwer(np.array([0.1, 0.5, 1]), quantile=1.1)

    with pytest.raises(ValueError):
        assert quantile_based_fwer(np.array([0.1, 0.5, 1]), quantile=-0.5)


def test_marginal_expectation_returns_all_results():
    # Just checking formats, i.e. no need for correlation.
    X = np.random.normal(0, 1, (1000, 3))
    Y = np.random.normal(0, 1, (1000, 1))

    model_all_features = create_linear_regressor()
    model_all_features.fit(X, Y)

    results = marginal_expectation(model_all_features.predict, X, X, [0], return_averaged_results=False)
    assert results.shape[0] == 1000
    assert results.shape[1] == 1000
    assert results.shape[2] == 1


def test_marginal_expectation_returns_reduced_results():
    # Just checking formats, i.e. no need for correlation.
    X = np.random.normal(0, 1, (1000, 3))
    Y = np.random.normal(0, 1, (1000, 1))

    model_all_features = create_linear_regressor()
    model_all_features.fit(X, Y)

    results = marginal_expectation(model_all_features.predict, X, X, [0], return_averaged_results=True)
    assert results.shape[0] == 1000
    assert results.shape[1] == 1


@flaky(max_runs=5)
def test_marginal_expectation_independent_continuous_linear():
    X = np.random.normal(0, 1, (1000, 3))
    Y = 3 * X[:, 0] + 2 * X[:, 1] - X[:, 2]
    Y = Y.reshape(-1)

    X0 = X[:, 0].reshape(-1, 1)
    X1 = X[:, 1].reshape(-1, 1)
    X2 = X[:, 2].reshape(-1, 1)
    X01 = X[:, :2]

    model_all_features = create_linear_regressor()
    model_all_features.fit(X, Y)
    model_feature_0 = create_linear_regressor()
    model_feature_0.fit(X0, Y)
    model_feature_1 = create_linear_regressor()
    model_feature_1.fit(X1, Y)
    model_feature_2 = create_linear_regressor()
    model_feature_2.fit(X2, Y)
    model_feature_01 = create_linear_regressor()
    model_feature_01.fit(X01, Y)

    assert np.mean(
        np.abs(
            model_feature_0.predict(X0).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X, [0], feature_perturbation="randomize_columns_independently"
            )
        )
    ) == approx(0.2, abs=1)
    assert np.mean(
        np.abs(
            model_feature_1.predict(X1).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X, [1], feature_perturbation="randomize_columns_independently"
            )
        )
    ) == approx(0.2, abs=1)
    assert np.mean(
        np.abs(
            model_feature_2.predict(X2).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X, [2], feature_perturbation="randomize_columns_independently"
            )
        )
    ) == approx(0.2, abs=1)
    assert np.mean(
        np.abs(
            model_feature_01.predict(X01).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X, [0, 1], feature_perturbation="randomize_columns_independently"
            )
        )
    ) == approx(0.2, abs=1)

    assert np.mean(
        np.abs(
            model_feature_0.predict(X0).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X, [0], feature_perturbation="randomize_columns_jointly"
            )
        )
    ) == approx(0.2, abs=1)
    assert np.mean(
        np.abs(
            model_feature_1.predict(X1).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X, [1], feature_perturbation="randomize_columns_jointly"
            )
        )
    ) == approx(0.2, abs=1)
    assert np.mean(
        np.abs(
            model_feature_2.predict(X2).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X, [2], feature_perturbation="randomize_columns_jointly"
            )
        )
    ) == approx(0.2, abs=1)
    assert np.mean(
        np.abs(
            model_feature_01.predict(X01).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X, [0, 1], feature_perturbation="randomize_columns_jointly"
            )
        )
    ) == approx(0.2, abs=1)


@flaky(max_runs=5)
def test_marginal_expectation_independent_categorical_linear():
    X = np.random.normal(0, 1, (1000, 3))
    Y = 3 * X[:, 0] + 2 * X[:, 1] - X[:, 2]
    Y = (Y <= 0).reshape(-1)

    X0 = X[:, 0].reshape(-1, 1)
    X1 = X[:, 1].reshape(-1, 1)
    X2 = X[:, 2].reshape(-1, 1)
    X01 = X[:, :2]

    model_all_features = create_logistic_regression_classifier()
    model_all_features.fit(X, Y)
    model_feature_0 = create_logistic_regression_classifier()
    model_feature_0.fit(X0, Y)
    model_feature_1 = create_logistic_regression_classifier()
    model_feature_1.fit(X1, Y)
    model_feature_2 = create_logistic_regression_classifier()
    model_feature_2.fit(X2, Y)
    model_feature_01 = create_logistic_regression_classifier()
    model_feature_01.fit(X01, Y)

    assert np.sum(
        np.mean(
            np.abs(
                model_feature_0.predict_probabilities(X0)
                - marginal_expectation(
                    model_all_features.predict_probabilities,
                    X,
                    X,
                    [0],
                    feature_perturbation="randomize_columns_independently",
                )
            ),
            axis=0,
        )
    ) == approx(0.2, abs=1)
    assert np.sum(
        np.mean(
            np.abs(
                model_feature_1.predict_probabilities(X1)
                - marginal_expectation(
                    model_all_features.predict_probabilities,
                    X,
                    X,
                    [1],
                    feature_perturbation="randomize_columns_independently",
                )
            ),
            axis=0,
        )
    ) == approx(0.2, abs=1)
    assert np.sum(
        np.mean(
            np.abs(
                model_feature_2.predict_probabilities(X2)
                - marginal_expectation(
                    model_all_features.predict_probabilities,
                    X,
                    X,
                    [2],
                    feature_perturbation="randomize_columns_independently",
                )
            ),
            axis=0,
        )
    ) == approx(0.2, abs=1)
    assert np.sum(
        np.mean(
            np.abs(
                model_feature_01.predict_probabilities(X01)
                - marginal_expectation(
                    model_all_features.predict_probabilities,
                    X,
                    X,
                    [0, 1],
                    feature_perturbation="randomize_columns_independently",
                )
            ),
            axis=0,
        )
    ) == approx(0.2, abs=1)


@flaky(max_runs=5)
def test_marginal_expectation_independent_continuous_nonlinear():
    X = np.random.normal(0, 1, (2000, 3))
    Y = (2 * X[:, 0] + X[:, 1]) ** 2 + X[:, 2]
    Y = Y.reshape(-1)

    X0 = X[:, 0].reshape(-1, 1)
    X1 = X[:, 1].reshape(-1, 1)
    X2 = X[:, 2].reshape(-1, 1)
    X01 = X[:, :2]

    model_all_features = create_hist_gradient_boost_regressor()
    model_all_features.fit(X, Y)
    model_feature_0 = create_hist_gradient_boost_regressor()
    model_feature_0.fit(X0, Y)
    model_feature_1 = create_hist_gradient_boost_regressor()
    model_feature_1.fit(X1, Y)
    model_feature_2 = create_hist_gradient_boost_regressor()
    model_feature_2.fit(X2, Y)
    model_feature_01 = create_hist_gradient_boost_regressor()
    model_feature_01.fit(X01, Y)

    assert np.mean(
        np.abs(
            model_feature_0.predict(X0[:100]).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X[:100], [0], feature_perturbation="randomize_columns_independently"
            )
        )
    ) == approx(0, abs=100)
    assert np.mean(
        np.abs(
            model_feature_1.predict(X1[:100]).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X[:100], [1], feature_perturbation="randomize_columns_independently"
            )
        )
    ) == approx(0, abs=100)
    assert np.mean(
        np.abs(
            model_feature_2.predict(X2[:100]).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X[:100], [2], feature_perturbation="randomize_columns_independently"
            )
        )
    ) == approx(0, abs=100)
    assert np.mean(
        np.abs(
            model_feature_01.predict(X01[:100]).reshape(-1, 1)
            - marginal_expectation(
                model_all_features.predict, X, X[:100], [0, 1], feature_perturbation="randomize_columns_independently"
            )
        )
    ) == approx(0, abs=100)


@flaky(max_runs=5)
def test_marginal_expectation_independent_categorical_nonlinear():
    X = np.random.normal(0, 1, (1000, 3))
    Y = (2 * X[:, 0] + X[:, 1]) ** 2 + X[:, 2]
    Y = (Y <= np.mean(Y)).reshape(-1)

    X0 = X[:, 0].reshape(-1, 1)
    X1 = X[:, 1].reshape(-1, 1)
    X2 = X[:, 2].reshape(-1, 1)
    X01 = X[:, :2]

    model_all_features = create_hist_gradient_boost_classifier()
    model_all_features.fit(X, Y)
    model_feature_0 = create_hist_gradient_boost_classifier()
    model_feature_0.fit(X0, Y)
    model_feature_1 = create_hist_gradient_boost_classifier()
    model_feature_1.fit(X1, Y)
    model_feature_2 = create_hist_gradient_boost_classifier()
    model_feature_2.fit(X2, Y)
    model_feature_01 = create_hist_gradient_boost_classifier()
    model_feature_01.fit(X01, Y)

    assert np.sum(
        np.mean(
            np.abs(
                model_feature_0.predict_probabilities(X0[:100])
                - marginal_expectation(
                    model_all_features.predict_probabilities,
                    X,
                    X[:100],
                    [0],
                    feature_perturbation="randomize_columns_independently",
                )
            ),
            axis=0,
        )
    ) == approx(0.2, abs=1)
    assert np.sum(
        np.mean(
            np.abs(
                model_feature_1.predict_probabilities(X1[:100])
                - marginal_expectation(
                    model_all_features.predict_probabilities,
                    X,
                    X[:100],
                    [1],
                    feature_perturbation="randomize_columns_independently",
                )
            ),
            axis=0,
        )
    ) == approx(0.2, abs=1)
    assert np.sum(
        np.mean(
            np.abs(
                model_feature_2.predict_probabilities(X2[:100])
                - marginal_expectation(
                    model_all_features.predict_probabilities,
                    X,
                    X[:100],
                    [2],
                    feature_perturbation="randomize_columns_independently",
                )
            ),
            axis=0,
        )
    ) == approx(0.2, abs=1)
    assert np.sum(
        np.mean(
            np.abs(
                model_feature_01.predict_probabilities(X01[:100])
                - marginal_expectation(
                    model_all_features.predict_probabilities,
                    X,
                    X[:100],
                    [0, 1],
                    feature_perturbation="randomize_columns_independently",
                )
            ),
            axis=0,
        )
    ) == approx(0.2, abs=1)


def test_given_different_batch_sizes_when_estimating_marginal_expectation_then_returns_expected_result():
    X = np.random.normal(0, 1, (34, 3))
    feature_samples = np.random.normal(0, 1, (123, 3))
    expected_non_aggregated = np.array([repmat(X[i, :], feature_samples.shape[0], 1) for i in range(X.shape[0])])

    def my_pred_func(X: np.ndarray) -> np.ndarray:
        return X.copy()

    assert marginal_expectation(my_pred_func, feature_samples, X, [0, 1, 2], max_batch_size=1) == approx(X)
    assert marginal_expectation(my_pred_func, feature_samples, X, [0, 1, 2], max_batch_size=10) == approx(X)
    assert marginal_expectation(my_pred_func, feature_samples, X, [0, 1, 2], max_batch_size=100) == approx(X)
    assert marginal_expectation(my_pred_func, feature_samples, X, [0, 1, 2], max_batch_size=1000) == approx(X)
    assert marginal_expectation(
        my_pred_func, feature_samples, X, [0, 1, 2], max_batch_size=feature_samples.shape[0]
    ) == approx(X)

    assert marginal_expectation(
        my_pred_func, feature_samples, X, [0, 1, 2], max_batch_size=1, return_averaged_results=False
    ) == approx(expected_non_aggregated)
    assert marginal_expectation(
        my_pred_func, feature_samples, X, [0, 1, 2], max_batch_size=10, return_averaged_results=False
    ) == approx(expected_non_aggregated)
    assert marginal_expectation(
        my_pred_func, feature_samples, X, [0, 1, 2], max_batch_size=100, return_averaged_results=False
    ) == approx(expected_non_aggregated)
    assert marginal_expectation(
        my_pred_func, feature_samples, X, [0, 1, 2], max_batch_size=1000, return_averaged_results=False
    ) == approx(expected_non_aggregated)
    assert marginal_expectation(
        my_pred_func,
        feature_samples,
        X,
        [0, 1, 2],
        max_batch_size=feature_samples.shape[0],
        return_averaged_results=False,
    ) == approx(expected_non_aggregated)


@flaky(max_runs=2)
def test_given_linear_dependent_data_when_estimate_ftest_pvalue_then_returns_expected_result():
    X_training = np.random.normal(0, 1, 1000)
    Y_training = X_training + np.random.normal(0, 0.05, 1000)

    X_test = np.random.normal(0, 1, 1000)
    Y_test = X_test + np.random.normal(0, 0.05, 1000)

    assert estimate_ftest_pvalue(X_training, np.array([]), Y_training, X_test, np.array([]), Y_test) < 0.05

    Y_training = np.random.normal(0, 0.05, 1000)
    Y_test = np.random.normal(0, 0.05, 1000)

    assert estimate_ftest_pvalue(X_training, np.array([]), Y_training, X_test, np.array([]), Y_test) >= 0.05


@flaky(max_runs=2)
def test_given_multivariate_dependent_data_when_estimate_ftest_pvalue_then_returns_expected_result():
    X1_training = np.random.normal(0, 1, 1000)
    X2_training = np.random.normal(0, 1, 1000)
    Y_training = X1_training + X2_training + np.random.normal(0, 0.05, 1000)

    X1_test = np.random.normal(0, 1, 1000)
    X2_test = np.random.normal(0, 1, 1000)
    Y_test = X1_test + X2_test + np.random.normal(0, 0.05, 1000)

    assert (
        estimate_ftest_pvalue(
            np.column_stack([X1_training, X2_training]),
            X1_training,
            Y_training,
            np.column_stack([X1_test, X2_test]),
            X1_test,
            Y_test,
        )
        < 0.05
    )

    Y_training = X1_training + np.random.normal(0, 0.05, 1000)
    Y_test = X1_test + np.random.normal(0, 0.05, 1000)

    assert (
        estimate_ftest_pvalue(
            np.column_stack([X1_training, X2_training]),
            X1_training,
            Y_training,
            np.column_stack([X1_test, X2_test]),
            X1_test,
            Y_test,
        )
        >= 0.05
    )



import numpy as np
import pandas as pd

from dowhy.gcm.util.general import has_categorical, is_categorical


def test_given_categorical_data_when_evaluating_is_categorical_then_returns_expected_result():
    assert is_categorical(np.array(["A", "B", "A"]))
    assert is_categorical(np.array([True, False, False]))
    assert is_categorical(pd.DataFrame({"X": [True, False, False]}).to_numpy())
    assert not is_categorical(np.array([1, 2, 3]))


def test_given_categorical_data_when_evaluating_has_categorical_then_returns_expected_result():
    assert has_categorical(np.array([["A", 2, 3], ["B", 4, 5]]))
    assert has_categorical(
        np.column_stack([np.array([True, False, False], dtype=object), np.array([1, 2, 3], dtype=object)])
    )
    assert has_categorical(pd.DataFrame({"X": [True, False, False], "Y": [1, 2, 3]}).to_numpy())
    assert not has_categorical(np.array([[1, 2, 3], [12.2, 2.3, 3.231]]))



import numpy as np
import pytest

from dowhy.gcm import confidence_intervals


def test_given_simple_counter_as_estimation_func_when_confidence_interval_then_returns_mean_and_interval_counter():
    i = 0.0

    def simple_counter():
        nonlocal i
        i += 1.0
        return {"X": i}

    median, interval = confidence_intervals(simple_counter, num_bootstrap_resamples=20)

    assert median["X"] == pytest.approx(10.5)
    assert np.allclose(interval["X"], [1.95, 19.05])



"""Functions and classes in this module should be considered experimental, meaning there might be breaking API changes
in the future.
"""

from typing import Any

import numpy as np
import sklearn
from packaging import version
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import PolynomialFeatures

if version.parse(sklearn.__version__) < version.parse("1.0"):
    from sklearn.experimental import enable_hist_gradient_boosting  # noqa

from sklearn.ensemble import (
    AdaBoostRegressor,
    ExtraTreesRegressor,
    HistGradientBoostingRegressor,
    RandomForestRegressor,
)
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.linear_model import ElasticNetCV, LassoCV, LassoLarsIC, LinearRegression, RidgeCV
from sklearn.neighbors import KNeighborsRegressor
from sklearn.svm import SVR

from dowhy.gcm.fcms import InvertibleFunction, PredictionModel
from dowhy.gcm.util.general import apply_one_hot_encoding, fit_one_hot_encoders, shape_into_2d


class SklearnRegressionModel(PredictionModel):
    """
    General wrapper class for sklearn models.
    """

    def __init__(self, sklearn_mdl: Any) -> None:
        self._sklearn_mdl = sklearn_mdl
        self._one_hot_encoders = {}

    def fit(self, X: np.ndarray, Y: np.ndarray) -> None:
        self._one_hot_encoders = fit_one_hot_encoders(X)
        X = apply_one_hot_encoding(X, self._one_hot_encoders)

        self._sklearn_mdl.fit(X=X, y=Y.squeeze())

    def predict(self, X: np.array) -> np.ndarray:
        return shape_into_2d(self._sklearn_mdl.predict(apply_one_hot_encoding(X, self._one_hot_encoders)))

    @property
    def sklearn_model(self) -> Any:
        return self._sklearn_mdl

    def clone(self):
        """
        Clones the prediction model using the same hyper parameters but not fitted.
        :return: An unfitted clone of the prediction model.
        """
        return SklearnRegressionModel(sklearn_mdl=sklearn.clone(self._sklearn_mdl))


def create_linear_regressor_with_given_parameters(
    coefficients: np.ndarray, intercept: float = 0, **kwargs
) -> SklearnRegressionModel:
    linear_model = LinearRegression(**kwargs)
    linear_model.coef_ = coefficients
    linear_model.intercept_ = intercept

    return SklearnRegressionModel(linear_model)


def create_linear_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(LinearRegression(**kwargs))


def create_ridge_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(RidgeCV(**kwargs))


def create_lasso_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(LassoCV(**kwargs))


def create_lasso_lars_ic_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(LassoLarsIC(**kwargs))


def create_elastic_net_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(ElasticNetCV(**kwargs))


def create_gaussian_process_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(GaussianProcessRegressor(**kwargs))


def create_support_vector_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(SVR(**kwargs))


def create_random_forest_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(RandomForestRegressor(**kwargs))


def create_hist_gradient_boost_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(HistGradientBoostingRegressor(**kwargs))


def create_extra_trees_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(ExtraTreesRegressor(**kwargs))


def create_knn_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(KNeighborsRegressor(**kwargs))


def create_ada_boost_regressor(**kwargs) -> SklearnRegressionModel:
    return SklearnRegressionModel(AdaBoostRegressor(**kwargs))


def create_polynom_regressor(degree: int = 3, **kwargs_linear_model) -> SklearnRegressionModel:
    return SklearnRegressionModel(
        make_pipeline(PolynomialFeatures(degree=degree, include_bias=False), LinearRegression(**kwargs_linear_model))
    )


class InvertibleIdentityFunction(InvertibleFunction):
    def evaluate(self, X: np.ndarray) -> np.ndarray:
        return X

    def evaluate_inverse(self, X: np.ndarray) -> np.ndarray:
        return X


class InvertibleExponentialFunction(InvertibleFunction):
    def evaluate(self, X: np.ndarray) -> np.ndarray:
        return np.exp(X)

    def evaluate_inverse(self, X: np.ndarray) -> np.ndarray:
        return np.log(X)


class InvertibleLogarithmicFunction(InvertibleFunction):
    def evaluate(self, X: np.ndarray) -> np.ndarray:
        return np.log(X)

    def evaluate_inverse(self, X: np.ndarray) -> np.ndarray:
        return np.exp(X)
